# encoding: utf-8

"""Unit-test suite for `pptx.oxml.shapes.picture` module."""

import pytest

from pptx.oxml.ns import nsdecls
from pptx.oxml.shapes.picture import CT_Picture


class DescribeCT_Picture(object):
    """Unit-test suite for `pptx.oxml.shapes.picture.CT_Picture` objects."""

    @pytest.mark.parametrize(
        "desc, xml_desc",
        (
            ("kittens.jpg", "kittens.jpg"),
            ("bits&bobs.png", "bits&amp;bobs.png"),
            ("img&.png", "img&amp;.png"),
            ("im<ag>e.png", "im&lt;ag&gt;e.png"),
        ),
    )
    def it_can_create_a_new_pic_element(self, desc, xml_desc):
        """`desc` attr (often filename) is XML-escaped to handle special characters.

        In particular, ampersand ('&'), less/greater-than ('</>') etc.
        """
        pic = CT_Picture.new_pic(
            shape_id=9, name="Picture 8", desc=desc, rId="rId42", x=1, y=2, cx=3, cy=4
        )

        assert pic.xml == (
            "<p:pic %s>\n"
            "  <p:nvPicPr>\n"
            '    <p:cNvPr id="9" name="Picture 8" descr="%s"/>\n'
            "    <p:cNvPicPr>\n"
            '      <a:picLocks noChangeAspect="1"/>\n'
            "    </p:cNvPicPr>\n"
            "    <p:nvPr/>\n"
            "  </p:nvPicPr>\n"
            "  <p:blipFill>\n"
            '    <a:blip r:embed="rId42"/>\n'
            "    <a:stretch>\n"
            "      <a:fillRect/>\n"
            "    </a:stretch>\n"
            "  </p:blipFill>\n"
            "  <p:spPr>\n"
            "    <a:xfrm>\n"
            '      <a:off x="1" y="2"/>\n'
            '      <a:ext cx="3" cy="4"/>\n'
            "    </a:xfrm>\n"
            '    <a:prstGeom prst="rect">\n'
            "      <a:avLst/>\n"
            "    </a:prstGeom>\n"
            "  </p:spPr>\n"
            "</p:pic>\n" % (nsdecls("a", "p", "r"), xml_desc)
        )

    def it_can_create_a_new_video_pic_element(self):
        pic = CT_Picture.new_video_pic(
            shape_id=42,
            shape_name="Media 41",
            video_rId="rId1",
            media_rId="rId2",
            poster_frame_rId="rId3",
            x=1,
            y=2,
            cx=3,
            cy=4,
        )

        assert pic.xml == (
            "<p:pic %s>\n"
            "  <p:nvPicPr>\n"
            '    <p:cNvPr id="42" name="Media 41">\n'
            '      <a:hlinkClick r:id="" action="ppaction://media"/>\n'
            "    </p:cNvPr>\n"
            "    <p:cNvPicPr>\n"
            '      <a:picLocks noChangeAspect="1"/>\n'
            "    </p:cNvPicPr>\n"
            "    <p:nvPr>\n"
            '      <a:videoFile r:link="rId1"/>\n'
            "      <p:extLst>\n"
            '        <p:ext uri="{DAA4B4D4-6D71-4841-9C94-3DE7FCFB9230}">\n'
            '          <p14:media xmlns:p14="http://schemas.microsoft.com/office/power'
            'point/2010/main" r:embed="rId2"/>\n'
            "        </p:ext>\n"
            "      </p:extLst>\n"
            "    </p:nvPr>\n"
            "  </p:nvPicPr>\n"
            "  <p:blipFill>\n"
            '    <a:blip r:embed="rId3"/>\n'
            "    <a:stretch>\n"
            "      <a:fillRect/>\n"
            "    </a:stretch>\n"
            "  </p:blipFill>\n"
            "  <p:spPr>\n"
            "    <a:xfrm>\n"
            '      <a:off x="1" y="2"/>\n'
            '      <a:ext cx="3" cy="4"/>\n'
            "    </a:xfrm>\n"
            '    <a:prstGeom prst="rect">\n'
            "      <a:avLst/>\n"
            "    </a:prstGeom>\n"
            "  </p:spPr>\n"
            "</p:pic>\n"
        ) % nsdecls("a", "p", "r")



# encoding: utf-8

"""Unit-test suite for `pptx.parts.image` module."""

import pytest

from pptx.compat import BytesIO
from pptx.package import Package
from pptx.parts.image import Image, ImagePart
from pptx.util import Emu

from ..unitutil.file import absjoin, test_file_dir
from ..unitutil.mock import (
    class_mock,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


images_pptx_path = absjoin(test_file_dir, "with_images.pptx")

test_image_path = absjoin(test_file_dir, "python-icon.jpeg")
test_eps_path = absjoin(test_file_dir, "cdw-logo.eps")
new_image_path = absjoin(test_file_dir, "monty-truth.png")


class DescribeImagePart(object):
    """Unit-test suite for `pptx.parts.image.ImagePart` objects."""

    def it_can_construct_from_an_image_object(self, request, image_):
        package_ = instance_mock(request, Package)
        _init_ = initializer_mock(request, ImagePart)
        partname_ = package_.next_image_partname.return_value

        image_part = ImagePart.new(package_, image_)

        package_.next_image_partname.assert_called_once_with(image_.ext)
        _init_.assert_called_once_with(
            image_part,
            partname_,
            image_.content_type,
            package_,
            image_.blob,
            image_.filename,
        )
        assert isinstance(image_part, ImagePart)

    def it_provides_access_to_its_image(self, request, image_):
        Image_ = class_mock(request, "pptx.parts.image.Image")
        Image_.return_value = image_
        property_mock(request, ImagePart, "desc", return_value="foobar.png")
        image_part = ImagePart(None, None, None, b"blob", None)

        image = image_part.image

        Image_.assert_called_once_with(b"blob", "foobar.png")
        assert image is image_

    @pytest.mark.parametrize(
        "width, height, expected_width, expected_height",
        (
            (None, None, Emu(2590800), Emu(2590800)),
            (1000, None, 1000, 1000),
            (None, 3000, 3000, 3000),
            (3337, 9999, 3337, 9999),
        ),
    )
    def it_can_scale_its_dimensions(
        self, width, height, expected_width, expected_height
    ):
        with open(test_image_path, "rb") as f:
            blob = f.read()
        image_part = ImagePart(None, None, None, blob)

        assert image_part.scale(width, height) == (expected_width, expected_height)

    def it_knows_its_pixel_dimensions_to_help(self):
        with open(test_image_path, "rb") as f:
            blob = f.read()
        image_part = ImagePart(None, None, None, blob)

        assert image_part._px_size == (204, 204)

    # fixture components ---------------------------------------------

    @pytest.fixture
    def image_(self, request):
        return instance_mock(request, Image)


class DescribeImage(object):
    """Unit-test suite for `pptx.parts.image.Image` objects."""

    def it_can_construct_from_a_path(self, from_blob_, image_):
        with open(test_image_path, "rb") as f:
            blob = f.read()
        from_blob_.return_value = image_

        image = Image.from_file(test_image_path)

        Image.from_blob.assert_called_once_with(blob, "python-icon.jpeg")
        assert image is image_

    def it_can_construct_from_a_stream(self, from_stream_fixture):
        image_file, blob, image_ = from_stream_fixture
        image = Image.from_file(image_file)
        Image.from_blob.assert_called_once_with(blob, None)
        assert image is image_

    def it_can_construct_from_a_blob(self, _init_):
        image = Image.from_blob(b"blob", "foo.png")

        _init_.assert_called_once_with(image, b"blob", "foo.png")
        assert isinstance(image, Image)

    def it_knows_its_blob(self, blob_fixture):
        image, expected_value = blob_fixture
        assert image.blob == expected_value

    def it_knows_its_content_type(self, content_type_fixture):
        image, expected_value = content_type_fixture
        assert image.content_type == expected_value

    def it_knows_its_canonical_filename_extension(self, ext_fixture):
        image, expected_value = ext_fixture
        assert image.ext == expected_value

    def it_knows_its_dpi(self, dpi_fixture):
        image, expected_value = dpi_fixture
        assert image.dpi == expected_value

    def it_knows_its_filename(self, filename_fixture):
        image, expected_value = filename_fixture
        assert image.filename == expected_value

    def it_knows_its_sha1_hash(self):
        image = Image(b"foobar", None)
        assert image.sha1 == "8843d7f92416211de9ebb963ff4ce28125932878"

    def it_knows_its_PIL_properties_to_help(self, pil_fixture):
        image, size, format, dpi = pil_fixture
        assert image.size == size
        assert image._format == format
        assert image.dpi == dpi
        assert image._pil_props == (format, size, None)

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def blob_fixture(self):
        blob = b"foobar"
        image = Image(blob, None)
        return image, blob

    @pytest.fixture(
        params=[
            ("BMP", "image/bmp"),
            ("GIF", "image/gif"),
            ("JPEG", "image/jpeg"),
            ("PNG", "image/png"),
            ("TIFF", "image/tiff"),
            ("WMF", "image/x-wmf"),
        ]
    )
    def content_type_fixture(self, request, _format_):
        format, expected_value = request.param
        image = Image(None, None)
        _format_.return_value = format
        return image, expected_value

    @pytest.fixture(
        params=[
            ((42, 24), (42, 24)),
            ((42.1, 23.6), (42, 24)),
            (None, (72, 72)),
            ((3047, 2388), (72, 72)),
            ("foobar", (72, 72)),
        ]
    )
    def dpi_fixture(self, request, _pil_props_):
        raw_dpi, expected_dpi = request.param
        image = Image(None, None)
        _pil_props_.return_value = (None, None, raw_dpi)
        return image, expected_dpi

    @pytest.fixture(
        params=[
            ("BMP", "bmp"),
            ("GIF", "gif"),
            ("JPEG", "jpg"),
            ("PNG", "png"),
            ("TIFF", "tiff"),
            ("WMF", "wmf"),
        ]
    )
    def ext_fixture(self, request, _format_):
        format, expected_value = request.param
        image = Image(None, None)
        _format_.return_value = format
        return image, expected_value

    @pytest.fixture(params=["foo.bar", None])
    def filename_fixture(self, request):
        filename = request.param
        image = Image(None, filename)
        return image, filename

    @pytest.fixture
    def from_stream_fixture(self, from_blob_, image_):
        with open(test_image_path, "rb") as f:
            blob = f.read()
            image_file = BytesIO(blob)
        from_blob_.return_value = image_
        return image_file, blob, image_

    @pytest.fixture
    def pil_fixture(self):
        with open(test_image_path, "rb") as f:
            blob = f.read()
        image = Image(blob, None)
        size = (204, 204)
        format = "JPEG"
        dpi = (72, 72)
        return image, size, format, dpi

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _format_(self, request):
        return property_mock(request, Image, "_format")

    @pytest.fixture
    def from_blob_(self, request):
        return method_mock(request, Image, "from_blob", autospec=False)

    @pytest.fixture
    def image_(self, request):
        return instance_mock(request, Image)

    @pytest.fixture
    def _init_(self, request):
        return initializer_mock(request, Image)

    @pytest.fixture
    def _pil_props_(self, request):
        return property_mock(request, Image, "_pil_props")



# encoding: utf-8

"""Unit-test suite for `pptx.parts.presentation` module."""

import pytest

from pptx.opc.constants import RELATIONSHIP_TYPE as RT
from pptx.opc.packuri import PackURI
from pptx.package import Package
from pptx.parts.coreprops import CorePropertiesPart
from pptx.parts.presentation import PresentationPart
from pptx.parts.slide import NotesMasterPart, SlideMasterPart, SlidePart
from pptx.presentation import Presentation
from pptx.slide import NotesMaster, Slide, SlideLayout, SlideMaster

from ..unitutil.cxml import element
from ..unitutil.mock import call, class_mock, instance_mock, method_mock, property_mock


class DescribePresentationPart(object):
    """Unit-test suite for `pptx.parts.presentation.PresentationPart` objects."""

    def it_provides_access_to_its_presentation(self, request):
        prs_ = instance_mock(request, Presentation)
        Presentation_ = class_mock(
            request, "pptx.parts.presentation.Presentation", return_value=prs_
        )
        prs_elm = element("p:presentation")
        prs_part = PresentationPart(None, None, None, prs_elm)

        prs = prs_part.presentation

        Presentation_.assert_called_once_with(prs_elm, prs_part)
        assert prs is prs_

    def it_provides_access_to_its_core_properties(self, request, package_):
        core_properties_ = instance_mock(request, CorePropertiesPart)
        package_.core_properties = core_properties_
        prs_part = PresentationPart(None, None, package_, None)

        assert prs_part.core_properties is core_properties_

    def it_provides_access_to_an_existing_notes_master_part(
        self, notes_master_part_, part_related_by_
    ):
        """This is the first of a two-part test to cover the existing notes master case.

        The notes master not-present case follows.
        """
        prs_part = PresentationPart(None, None, None, None)
        part_related_by_.return_value = notes_master_part_

        notes_master_part = prs_part.notes_master_part

        prs_part.part_related_by.assert_called_once_with(prs_part, RT.NOTES_MASTER)
        assert notes_master_part is notes_master_part_

    def but_it_adds_a_notes_master_part_when_needed(
        self, request, package_, notes_master_part_, part_related_by_, relate_to_
    ):
        """This is the second of a two-part test to cover notes-master-not-present case.

        The notes master present case is just above.
        """
        NotesMasterPart_ = class_mock(
            request, "pptx.parts.presentation.NotesMasterPart"
        )
        NotesMasterPart_.create_default.return_value = notes_master_part_
        part_related_by_.side_effect = KeyError
        prs_part = PresentationPart(None, None, package_, None)

        notes_master_part = prs_part.notes_master_part

        NotesMasterPart_.create_default.assert_called_once_with(package_)
        relate_to_.assert_called_once_with(
            prs_part, notes_master_part_, RT.NOTES_MASTER
        )
        assert notes_master_part is notes_master_part_

    def it_provides_access_to_its_notes_master(self, request, notes_master_part_):
        notes_master_ = instance_mock(request, NotesMaster)
        property_mock(
            request,
            PresentationPart,
            "notes_master_part",
            return_value=notes_master_part_,
        )
        notes_master_part_.notes_master = notes_master_
        prs_part = PresentationPart(None, None, None, None)

        assert prs_part.notes_master is notes_master_

    def it_provides_access_to_a_related_slide(self, request, slide_, related_part_):
        slide_part_ = instance_mock(request, SlidePart, slide=slide_)
        related_part_.return_value = slide_part_
        prs_part = PresentationPart(None, None, None, None)

        slide = prs_part.related_slide("rId42")

        related_part_.assert_called_once_with(prs_part, "rId42")
        assert slide is slide_

    def it_provides_access_to_a_related_master(
        self, request, slide_master_, related_part_
    ):
        slide_master_part_ = instance_mock(
            request, SlideMasterPart, slide_master=slide_master_
        )
        related_part_.return_value = slide_master_part_
        prs_part = PresentationPart(None, None, None, None)

        slide_master = prs_part.related_slide_master("rId42")

        related_part_.assert_called_once_with(prs_part, "rId42")
        assert slide_master is slide_master_

    def it_can_rename_related_slide_parts(self, request, related_part_):
        rIds = tuple("rId%d" % n for n in range(5, 0, -1))
        slide_parts = tuple(instance_mock(request, SlidePart) for _ in range(5))
        related_part_.side_effect = iter(slide_parts)
        prs_part = PresentationPart(None, None, None, None)

        prs_part.rename_slide_parts(rIds)

        assert related_part_.call_args_list == [call(prs_part, rId) for rId in rIds]
        assert [s.partname for s in slide_parts] == [
            PackURI("/ppt/slides/slide%d.xml" % (i + 1)) for i in range(len(rIds))
        ]

    def it_can_save_the_package_to_a_file(self, package_):
        PresentationPart(None, None, package_, None).save("prs.pptx")
        package_.save.assert_called_once_with("prs.pptx")

    def it_can_add_a_new_slide(
        self, request, package_, slide_part_, slide_, relate_to_
    ):
        slide_layout_ = instance_mock(request, SlideLayout)
        partname = PackURI("/ppt/slides/slide9.xml")
        property_mock(
            request, PresentationPart, "_next_slide_partname", return_value=partname
        )
        SlidePart_ = class_mock(request, "pptx.parts.presentation.SlidePart")
        SlidePart_.new.return_value = slide_part_
        relate_to_.return_value = "rId42"
        slide_layout_part_ = slide_layout_.part
        slide_part_.slide = slide_
        prs_part = PresentationPart(None, None, package_, None)

        rId, slide = prs_part.add_slide(slide_layout_)

        SlidePart_.new.assert_called_once_with(partname, package_, slide_layout_part_)
        prs_part.relate_to.assert_called_once_with(prs_part, slide_part_, RT.SLIDE)
        assert rId == "rId42"
        assert slide is slide_

    def it_finds_the_slide_id_of_a_slide_part(self, slide_part_, related_part_):
        prs_elm = element(
            "p:presentation/p:sldIdLst/(p:sldId{r:id=a,id=256},p:sldId{r:id="
            "b,id=257},p:sldId{r:id=c,id=258})"
        )
        related_part_.side_effect = iter((None, slide_part_, None))
        prs_part = PresentationPart(None, None, None, prs_elm)

        _slide_id = prs_part.slide_id(slide_part_)

        assert related_part_.call_args_list == [
            call(prs_part, "a"),
            call(prs_part, "b"),
        ]
        assert _slide_id == 257

    def it_raises_on_slide_id_not_found(self, slide_part_, related_part_):
        prs_elm = element(
            "p:presentation/p:sldIdLst/(p:sldId{r:id=a,id=256},p:sldId{r:id="
            "b,id=257},p:sldId{r:id=c,id=258})"
        )
        related_part_.return_value = "not the slide you're looking for"
        prs_part = PresentationPart(None, None, None, prs_elm)

        with pytest.raises(ValueError):
            prs_part.slide_id(slide_part_)

    @pytest.mark.parametrize("is_present", (True, False))
    def it_finds_a_slide_by_slide_id(
        self, is_present, slide_, slide_part_, related_part_
    ):
        prs_elm = element(
            "p:presentation/p:sldIdLst/(p:sldId{r:id=a,id=256},p:sldId{r:id="
            "b,id=257},p:sldId{r:id=c,id=258})"
        )
        slide_id = 257 if is_present else 666
        expected_value = slide_ if is_present else None
        related_part_.return_value = slide_part_
        slide_part_.slide = slide_
        prs_part = PresentationPart(None, None, None, prs_elm)

        slide = prs_part.get_slide(slide_id)

        assert slide == expected_value

    def it_knows_the_next_slide_partname_to_help(self):
        prs_elm = element("p:presentation/p:sldIdLst/(p:sldId,p:sldId)")
        prs_part = PresentationPart(None, None, None, prs_elm)

        assert prs_part._next_slide_partname == PackURI("/ppt/slides/slide3.xml")

    # fixture components ---------------------------------------------

    @pytest.fixture
    def notes_master_part_(self, request):
        return instance_mock(request, NotesMasterPart)

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, Package)

    @pytest.fixture
    def part_related_by_(self, request):
        return method_mock(request, PresentationPart, "part_related_by")

    @pytest.fixture
    def relate_to_(self, request):
        return method_mock(request, PresentationPart, "relate_to")

    @pytest.fixture
    def related_part_(self, request):
        return method_mock(request, PresentationPart, "related_part")

    @pytest.fixture
    def slide_(self, request):
        return instance_mock(request, Slide)

    @pytest.fixture
    def slide_master_(self, request):
        return instance_mock(request, SlideMaster)

    @pytest.fixture
    def slide_part_(self, request):
        return instance_mock(request, SlidePart)



# encoding: utf-8

"""Presentation part, the main part in a .pptx package."""

from pptx.opc.constants import RELATIONSHIP_TYPE as RT
from pptx.opc.package import XmlPart
from pptx.opc.packuri import PackURI
from pptx.parts.slide import NotesMasterPart, SlidePart
from pptx.presentation import Presentation
from pptx.util import lazyproperty


class PresentationPart(XmlPart):
    """Top level class in object model.

    Represents the contents of the /ppt directory of a .pptx file.
    """

    def add_slide(self, slide_layout):
        """
        Return an (rId, slide) pair of a newly created blank slide that
        inherits appearance from *slide_layout*.
        """
        partname = self._next_slide_partname
        slide_layout_part = slide_layout.part
        slide_part = SlidePart.new(partname, self.package, slide_layout_part)
        rId = self.relate_to(slide_part, RT.SLIDE)
        return rId, slide_part.slide

    @property
    def core_properties(self):
        """
        A |CoreProperties| object providing read/write access to the core
        properties of this presentation.
        """
        return self.package.core_properties

    def get_slide(self, slide_id):
        """Return optional related |Slide| object identified by `slide_id`.

        Returns |None| if no slide with `slide_id` is related to this presentation.
        """
        for sldId in self._element.sldIdLst:
            if sldId.id == slide_id:
                return self.related_part(sldId.rId).slide
        return None

    @lazyproperty
    def notes_master(self):
        """
        Return the |NotesMaster| object for this presentation. If the
        presentation does not have a notes master, one is created from
        a default template. The same single instance is returned on each
        call.
        """
        return self.notes_master_part.notes_master

    @lazyproperty
    def notes_master_part(self):
        """
        Return the |NotesMasterPart| object for this presentation. If the
        presentation does not have a notes master, one is created from
        a default template. The same single instance is returned on each
        call.
        """
        try:
            return self.part_related_by(RT.NOTES_MASTER)
        except KeyError:
            notes_master_part = NotesMasterPart.create_default(self.package)
            self.relate_to(notes_master_part, RT.NOTES_MASTER)
            return notes_master_part

    @lazyproperty
    def presentation(self):
        """
        A |Presentation| object providing access to the content of this
        presentation.
        """
        return Presentation(self._element, self)

    def related_slide(self, rId):
        """Return |Slide| object for related |SlidePart| related by `rId`."""
        return self.related_part(rId).slide

    def related_slide_master(self, rId):
        """Return |SlideMaster| object for |SlideMasterPart| related by `rId`."""
        return self.related_part(rId).slide_master

    def rename_slide_parts(self, rIds):
        """Assign incrementing partnames to the slide parts identified by `rIds`.

        Partnames are like `/ppt/slides/slide9.xml` and are assigned in the order their
        id appears in the `rIds` sequence. The name portion is always ``slide``. The
        number part forms a continuous sequence starting at 1 (e.g. 1, 2, ... 10, ...).
        The extension is always ``.xml``.
        """
        for idx, rId in enumerate(rIds):
            slide_part = self.related_part(rId)
            slide_part.partname = PackURI("/ppt/slides/slide%d.xml" % (idx + 1))

    def save(self, path_or_stream):
        """Save this presentation package to `path_or_stream`.

        `path_or_stream` can be either a path to a filesystem location (a string) or a
        file-like object.
        """
        self.package.save(path_or_stream)

    def slide_id(self, slide_part):
        """Return the slide-id associated with `slide_part`."""
        for sldId in self._element.sldIdLst:
            if self.related_part(sldId.rId) is slide_part:
                return sldId.id
        raise ValueError("matching slide_part not found")

    @property
    def _next_slide_partname(self):
        """Return |PackURI| instance containing next available slide partname."""
        sldIdLst = self._element.get_or_add_sldIdLst()
        partname_str = "/ppt/slides/slide%d.xml" % (len(sldIdLst) + 1)
        return PackURI(partname_str)



# encoding: utf-8

"""Unit-test suite for `pptx.opc.package` module."""

import collections
import io
import itertools

import pytest

from pptx.opc.constants import (
    CONTENT_TYPE as CT,
    RELATIONSHIP_TARGET_MODE as RTM,
    RELATIONSHIP_TYPE as RT,
)
from pptx.opc.oxml import CT_Relationship, CT_Relationships
from pptx.opc.package import (
    OpcPackage,
    Part,
    PartFactory,
    XmlPart,
    _ContentTypeMap,
    _PackageLoader,
    _RelatableMixin,
    _Relationship,
    _Relationships,
)
from pptx.opc.packuri import PACKAGE_URI, PackURI
from pptx.oxml import parse_xml
from pptx.parts.presentation import PresentationPart

from ..unitutil.cxml import element
from ..unitutil.file import absjoin, snippet_bytes, testfile_bytes, test_file_dir
from ..unitutil.mock import (
    ANY,
    call,
    class_mock,
    function_mock,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class Describe_RelatableMixin(object):
    """Unit-test suite for `pptx.opc.package._RelatableMixin`.

    This mixin is used for both OpcPackage and Part because both a package and a part
    can have relationships to target parts.
    """

    def it_can_find_a_part_related_by_reltype(self, _rels_prop_, relationships_, part_):
        relationships_.part_with_reltype.return_value = part_
        _rels_prop_.return_value = relationships_
        mixin = _RelatableMixin()

        related_part = mixin.part_related_by(RT.CHART)

        relationships_.part_with_reltype.assert_called_once_with(RT.CHART)
        assert related_part is part_

    def it_can_establish_a_relationship_to_another_part(
        self, _rels_prop_, relationships_, part_
    ):
        relationships_.get_or_add.return_value = "rId42"
        _rels_prop_.return_value = relationships_
        mixin = _RelatableMixin()

        rId = mixin.relate_to(part_, RT.SLIDE)

        relationships_.get_or_add.assert_called_once_with(RT.SLIDE, part_)
        assert rId == "rId42"

    def and_it_can_establish_a_relationship_to_an_external_link(
        self, request, _rels_prop_, relationships_
    ):
        relationships_.get_or_add_ext_rel.return_value = "rId24"
        _rels_prop_.return_value = relationships_
        mixin = _RelatableMixin()

        rId = mixin.relate_to("http://url", RT.HYPERLINK, is_external=True)

        relationships_.get_or_add_ext_rel.assert_called_once_with(
            RT.HYPERLINK, "http://url"
        )
        assert rId == "rId24"

    def it_can_find_a_related_part_by_rId(
        self, request, _rels_prop_, relationships_, relationship_, part_
    ):
        _rels_prop_.return_value = relationships_
        relationships_.__getitem__.return_value = relationship_
        relationship_.target_part = part_
        mixin = _RelatableMixin()

        related_part = mixin.related_part("rId17")

        relationships_.__getitem__.assert_called_once_with("rId17")
        assert related_part is part_

    def it_can_find_a_target_ref_URI_by_rId(
        self, request, _rels_prop_, relationships_, relationship_
    ):
        _rels_prop_.return_value = relationships_
        relationships_.__getitem__.return_value = relationship_
        relationship_.target_ref = "http://url"
        mixin = _RelatableMixin()

        target_ref = mixin.target_ref("rId9")

        relationships_.__getitem__.assert_called_once_with("rId9")
        assert target_ref == "http://url"

    # fixture components -----------------------------------

    @pytest.fixture
    def part_(self, request):
        return instance_mock(request, Part)

    @pytest.fixture
    def relationship_(self, request):
        return instance_mock(request, _Relationship)

    @pytest.fixture
    def relationships_(self, request):
        return instance_mock(request, _Relationships)

    @pytest.fixture
    def _rels_prop_(self, request):
        return property_mock(request, _RelatableMixin, "_rels")


class DescribeOpcPackage(object):
    """Unit-test suite for `pptx.opc.package.OpcPackage` objects."""

    def it_can_open_a_pkg_file(self, request):
        package_ = instance_mock(request, OpcPackage)
        _init_ = initializer_mock(request, OpcPackage)
        _load_ = method_mock(request, OpcPackage, "_load", return_value=package_)

        package = OpcPackage.open("package.pptx")

        _init_.assert_called_once_with(ANY, "package.pptx")
        _load_.assert_called_once_with(ANY)
        assert package is package_

    def it_can_drop_a_relationship(self, _rels_prop_, relationships_):
        _rels_prop_.return_value = relationships_

        OpcPackage(None).drop_rel("rId42")

        relationships_.pop.assert_called_once_with("rId42")

    def it_can_iterate_over_its_parts(self, request):
        part_, part_2_ = [
            instance_mock(request, Part, name="part_%d" % i) for i in range(2)
        ]
        rels_iter = (
            instance_mock(
                request, _Relationship, is_external=is_external, target_part=target
            )
            for is_external, target in (
                (True, "http://some/url/"),
                (False, part_),
                (False, part_),
                (False, part_2_),
                (False, part_),
                (False, part_2_),
            )
        )
        method_mock(request, OpcPackage, "iter_rels", return_value=rels_iter)
        package = OpcPackage(None)

        assert list(package.iter_parts()) == [part_, part_2_]

    def it_can_iterate_over_its_relationships(self, request, _rels_prop_):
        """
        +----------+          +--------+
        | pkg_rels |-- r0 --> | part_0 |
        +----------+          +--------+
             |     |            |    ^
          r2 |     | r1      r3 |    | r4
             |     |            |    |
             v     |            v    |
         external  |          +--------+
                   +--------> | part_1 |
                              +--------+
        """
        part_0_, part_1_ = [
            instance_mock(request, Part, name="part_%d" % i) for i in range(2)
        ]
        rels = tuple(
            instance_mock(
                request,
                _Relationship,
                name="r%d" % i,
                is_external=ext,
                target_part=part,
            )
            for i, (ext, part) in enumerate(
                (
                    (False, part_0_),
                    (False, part_1_),
                    (True, None),
                    (False, part_1_),
                    (False, part_0_),
                )
            )
        )
        _rels_prop_.return_value = rels[:3]
        part_0_.rels = rels[3:4]
        part_1_.rels = rels[4:]
        package = OpcPackage(None)

        assert tuple(package.iter_rels()) == (
            rels[0],
            rels[3],
            rels[4],
            rels[1],
            rels[2],
        )

    def it_provides_access_to_the_main_document_part(self, request):
        presentation_part_ = instance_mock(request, PresentationPart)
        part_related_by_ = method_mock(
            request, OpcPackage, "part_related_by", return_value=presentation_part_
        )
        package = OpcPackage(None)

        presentation_part = package.main_document_part

        part_related_by_.assert_called_once_with(package, RT.OFFICE_DOCUMENT)
        assert presentation_part is presentation_part_

    @pytest.mark.parametrize(
        "ns, expected_n",
        (((), 1), ((1,), 2), ((1, 2), 3), ((2, 4), 3), ((1, 4), 3)),
    )
    def it_can_find_the_next_available_partname(self, request, ns, expected_n):
        tmpl = "/x%d.xml"
        method_mock(
            request,
            OpcPackage,
            "iter_parts",
            return_value=(instance_mock(request, Part, partname=tmpl % n) for n in ns),
        )
        next_partname = tmpl % expected_n
        PackURI_ = class_mock(
            request, "pptx.opc.package.PackURI", return_value=PackURI(next_partname)
        )
        package = OpcPackage(None)

        partname = package.next_partname(tmpl)

        PackURI_.assert_called_once_with(next_partname)
        assert partname == next_partname

    def it_can_save_to_a_pkg_file(self, request, _rels_prop_, relationships_):
        _rels_prop_.return_value = relationships_
        parts_ = tuple(instance_mock(request, Part) for _ in range(3))
        method_mock(request, OpcPackage, "iter_parts", return_value=iter(parts_))
        PackageWriter_ = class_mock(request, "pptx.opc.package.PackageWriter")
        package = OpcPackage(None)

        package.save("prs.pptx")

        PackageWriter_.write.assert_called_once_with("prs.pptx", relationships_, parts_)

    def it_loads_the_pkg_file_to_help(self, request, _rels_prop_, relationships_):
        _PackageLoader_ = class_mock(request, "pptx.opc.package._PackageLoader")
        _PackageLoader_.load.return_value = "pkg-rels-xml", {"partname": "part"}
        _rels_prop_.return_value = relationships_
        package = OpcPackage("prs.pptx")

        return_value = package._load()

        _PackageLoader_.load.assert_called_once_with("prs.pptx", package)
        relationships_.load_from_xml.assert_called_once_with(
            PACKAGE_URI, "pkg-rels-xml", {"partname": "part"}
        )
        assert return_value is package

    def it_constructs_its_relationships_object_to_help(self, request, relationships_):
        _Relationships_ = class_mock(
            request, "pptx.opc.package._Relationships", return_value=relationships_
        )
        package = OpcPackage(None)

        rels = package._rels

        _Relationships_.assert_called_once_with(PACKAGE_URI.baseURI)
        assert rels is relationships_

    # fixture components -----------------------------------

    @pytest.fixture
    def relationships_(self, request):
        return instance_mock(request, _Relationships)

    @pytest.fixture
    def _rels_prop_(self, request):
        return property_mock(request, OpcPackage, "_rels")


class Describe_PackageLoader(object):
    """Unit-test suite for `pptx.opc.package._PackageLoader` objects."""

    def it_provides_a_load_interface_classmethod(self, request, package_):
        _init_ = initializer_mock(request, _PackageLoader)
        pkg_xml_rels_ = element("r:Relationships")
        _load_ = method_mock(
            request,
            _PackageLoader,
            "_load",
            return_value=(pkg_xml_rels_, {"partname": "part"}),
        )

        pkg_xml_rels, parts = _PackageLoader.load("prs.pptx", package_)

        _init_.assert_called_once_with(ANY, "prs.pptx", package_)
        _load_.assert_called_once_with(ANY)
        assert pkg_xml_rels is pkg_xml_rels_
        assert parts == {"partname": "part"}

    def it_loads_the_package_to_help(self, request, _xml_rels_prop_):
        parts_ = {
            "partname_%d" % n: instance_mock(request, Part, partname="partname_%d" % n)
            for n in range(1, 4)
        }
        property_mock(request, _PackageLoader, "_parts", return_value=parts_)
        rels_ = dict(
            itertools.chain(
                (("/", instance_mock(request, _Relationships)),),
                (
                    ("partname_%d" % n, instance_mock(request, _Relationships))
                    for n in range(1, 4)
                ),
            )
        )
        _xml_rels_prop_.return_value = rels_
        package_loader = _PackageLoader(None, None)

        pkg_xml_rels, parts = package_loader._load()

        for part_ in parts_.values():
            part_.load_rels_from_xml.assert_called_once_with(
                rels_[part_.partname], parts_
            )
        assert pkg_xml_rels is rels_["/"]
        assert parts is parts_

    def it_loads_the_xml_relationships_from_the_package_to_help(self, request):
        pkg_xml_rels = parse_xml(snippet_bytes("package-rels-xml"))
        prs_xml_rels = parse_xml(snippet_bytes("presentation-rels-xml"))
        slide_xml_rels = CT_Relationships.new()
        thumbnail_xml_rels = CT_Relationships.new()
        core_xml_rels = CT_Relationships.new()
        _xml_rels_for_ = method_mock(
            request,
            _PackageLoader,
            "_xml_rels_for",
            side_effect=iter(
                (
                    pkg_xml_rels,
                    prs_xml_rels,
                    slide_xml_rels,
                    thumbnail_xml_rels,
                    core_xml_rels,
                )
            ),
        )
        package_loader = _PackageLoader(None, None)

        xml_rels = package_loader._xml_rels

        # print(f"{_xml_rels_for_.call_args_list=}")
        assert _xml_rels_for_.call_args_list == [
            call(package_loader, "/"),
            call(package_loader, "/ppt/presentation.xml"),
            call(package_loader, "/ppt/slides/slide1.xml"),
            call(package_loader, "/docProps/thumbnail.jpeg"),
            call(package_loader, "/docProps/core.xml"),
        ]
        assert xml_rels == {
            "/": pkg_xml_rels,
            "/ppt/presentation.xml": prs_xml_rels,
            "/ppt/slides/slide1.xml": slide_xml_rels,
            "/docProps/thumbnail.jpeg": thumbnail_xml_rels,
            "/docProps/core.xml": core_xml_rels,
        }

    # fixture components -----------------------------------

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, OpcPackage)

    @pytest.fixture
    def _xml_rels_prop_(self, request):
        return property_mock(request, _PackageLoader, "_xml_rels")


class DescribePart(object):
    """Unit-test suite for `pptx.opc.package.Part` objects."""

    def it_can_be_constructed_by_PartFactory(self, request, package_):
        partname_ = instance_mock(request, PackURI)
        _init_ = initializer_mock(request, Part)

        part = Part.load(partname_, CT.PML_SLIDE, package_, b"blob")

        _init_.assert_called_once_with(part, partname_, CT.PML_SLIDE, package_, b"blob")
        assert isinstance(part, Part)

    def it_uses_the_load_blob_as_its_blob(self):
        assert Part(None, None, None, b"blob").blob == b"blob"

    def it_can_change_its_blob(self):
        part = Part(None, None, None, b"old-blob")
        part.blob = b"new-blob"
        assert part.blob == b"new-blob"

    def it_knows_its_content_type(self):
        assert Part(None, CT.PML_SLIDE, None).content_type == CT.PML_SLIDE

    @pytest.mark.parametrize("ref_count, calls", ((2, []), (1, [call("rId42")])))
    def it_can_drop_a_relationship(self, request, relationships_, ref_count, calls):
        _rel_ref_count_ = method_mock(
            request, Part, "_rel_ref_count", return_value=ref_count
        )
        property_mock(request, Part, "_rels", return_value=relationships_)
        part = Part(None, None, None)

        part.drop_rel("rId42")

        _rel_ref_count_.assert_called_once_with(part, "rId42")
        assert relationships_.pop.call_args_list == calls

    def it_knows_the_package_it_belongs_to(self, package_):
        assert Part(None, None, package_).package is package_

    def it_knows_its_partname(self):
        assert Part(PackURI("/part/name"), None, None).partname == PackURI("/part/name")

    def it_can_change_its_partname(self):
        part = Part(PackURI("/old/part/name"), None, None)
        part.partname = PackURI("/new/part/name")
        assert part.partname == PackURI("/new/part/name")

    def it_provides_access_to_its_relationships_for_traversal(
        self, request, relationships_
    ):
        property_mock(request, Part, "_rels", return_value=relationships_)
        assert Part(None, None, None).rels is relationships_

    def it_can_load_a_blob_from_a_file_path_to_help(self):
        path = absjoin(test_file_dir, "minimal.pptx")
        with open(path, "rb") as f:
            file_bytes = f.read()
        part = Part(None, None, None, None)

        assert part._blob_from_file(path) == file_bytes

    def it_can_load_a_blob_from_a_file_like_object_to_help(self):
        part = Part(None, None, None, None)
        assert part._blob_from_file(io.BytesIO(b"012345")) == b"012345"

    def it_constructs_its_relationships_object_to_help(self, request, relationships_):
        _Relationships_ = class_mock(
            request, "pptx.opc.package._Relationships", return_value=relationships_
        )
        part = Part(PackURI("/ppt/slides/slide1.xml"), None, None)

        rels = part._rels

        _Relationships_.assert_called_once_with("/ppt/slides")
        assert rels is relationships_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, OpcPackage)

    @pytest.fixture
    def relationships_(self, request):
        return instance_mock(request, _Relationships)


class DescribeXmlPart(object):
    """Unit-test suite for `pptx.opc.package.XmlPart` objects."""

    def it_can_be_constructed_by_PartFactory(self, request):
        partname = PackURI("/ppt/slides/slide1.xml")
        element_ = element("p:sld")
        package_ = instance_mock(request, OpcPackage)
        parse_xml_ = function_mock(
            request, "pptx.opc.package.parse_xml", return_value=element_
        )
        _init_ = initializer_mock(request, XmlPart)

        part = XmlPart.load(partname, CT.PML_SLIDE, package_, b"blob")

        parse_xml_.assert_called_once_with(b"blob")
        _init_.assert_called_once_with(part, partname, CT.PML_SLIDE, package_, element_)
        assert isinstance(part, XmlPart)

    def it_can_serialize_to_xml(self, request):
        element_ = element("p:sld")
        serialize_part_xml_ = function_mock(
            request, "pptx.opc.package.serialize_part_xml"
        )
        xml_part = XmlPart(None, None, None, element_)

        blob = xml_part.blob

        serialize_part_xml_.assert_called_once_with(element_)
        assert blob is serialize_part_xml_.return_value

    def it_knows_it_is_the_part_for_its_child_objects(self):
        xml_part = XmlPart(None, None, None, None)
        assert xml_part.part is xml_part


class DescribePartFactory(object):
    """Unit-test suite for `pptx.opc.package.PartFactory` objects."""

    def it_constructs_custom_part_type_for_registered_content_types(
        self, request, package_, part_
    ):
        SlidePart_ = class_mock(request, "pptx.opc.package.XmlPart")
        SlidePart_.load.return_value = part_
        partname = PackURI("/ppt/slides/slide7.xml")
        PartFactory.part_type_for[CT.PML_SLIDE] = SlidePart_

        part = PartFactory(partname, CT.PML_SLIDE, package_, b"blob")

        SlidePart_.load.assert_called_once_with(
            partname, CT.PML_SLIDE, package_, b"blob"
        )
        assert part is part_

    def it_constructs_part_using_default_class_when_no_custom_registered(
        self, request, package_, part_
    ):
        Part_ = class_mock(request, "pptx.opc.package.Part")
        Part_.load.return_value = part_
        partname = PackURI("/bar/foo.xml")

        part = PartFactory(partname, CT.OFC_VML_DRAWING, package_, b"blob")

        Part_.load.assert_called_once_with(
            partname, CT.OFC_VML_DRAWING, package_, b"blob"
        )
        assert part is part_

    # fixtures components ----------------------------------

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, OpcPackage)

    @pytest.fixture
    def part_(self, request):
        return instance_mock(request, Part)


class Describe_ContentTypeMap(object):
    """Unit-test suite for `pptx.opc.package._ContentTypeMap` objects."""

    def it_can_construct_from_content_types_xml(self, request):
        _init_ = initializer_mock(request, _ContentTypeMap)
        content_types_xml = (
            '<Types xmlns="http://schemas.openxmlformats.org/package/2006/content-ty'
            'pes">\n'
            '  <Default Extension="xml" ContentType="application/xml"/>\n'
            '  <Default Extension="PNG" ContentType="image/png"/>\n'
            '  <Override PartName="/ppt/presentation.xml" ContentType="application/v'
            'nd.openxmlformats-officedocument.presentationml.presentation.main+xml"/'
            ">\n"
            "</Types>\n"
        )

        ct_map = _ContentTypeMap.from_xml(content_types_xml)

        _init_.assert_called_once_with(
            ct_map,
            {"/ppt/presentation.xml": CT.PML_PRESENTATION_MAIN},
            {"png": CT.PNG, "xml": CT.XML},
        )

    @pytest.mark.parametrize(
        "partname, expected_value",
        (
            ("/docProps/core.xml", CT.OPC_CORE_PROPERTIES),
            ("/ppt/presentation.xml", CT.PML_PRESENTATION_MAIN),
            ("/PPT/Presentation.XML", CT.PML_PRESENTATION_MAIN),
            ("/ppt/viewprops.xml", CT.PML_VIEW_PROPS),
        ),
    )
    def it_matches_an_override_on_case_insensitive_partname(
        self, content_type_map, partname, expected_value
    ):
        assert content_type_map[PackURI(partname)] == expected_value

    @pytest.mark.parametrize(
        "partname, expected_value",
        (
            ("/foo/bar.xml", CT.XML),
            ("/FOO/BAR.Rels", CT.OPC_RELATIONSHIPS),
            ("/foo/bar.jpeg", CT.JPEG),
        ),
    )
    def it_falls_back_to_case_insensitive_extension_default_match(
        self, content_type_map, partname, expected_value
    ):
        assert content_type_map[PackURI(partname)] == expected_value

    def it_raises_KeyError_on_partname_not_found(self, content_type_map):
        with pytest.raises(KeyError) as e:
            content_type_map[PackURI("/!blat/rhumba.1x&")]
        assert str(e.value) == (
            "\"no content-type for partname '/!blat/rhumba.1x&' "
            'in [Content_Types].xml"'
        )

    def it_raises_TypeError_on_key_not_instance_of_PackURI(self, content_type_map):
        with pytest.raises(TypeError) as e:
            content_type_map["/part/name1.xml"]
        assert str(e.value) == "_ContentTypeMap key must be <type 'PackURI'>, got str"

    # fixtures ---------------------------------------------

    @pytest.fixture(scope="class")
    def content_type_map(self):
        return _ContentTypeMap.from_xml(
            testfile_bytes("expanded_pptx", "[Content_Types].xml")
        )


class Describe_Relationships(object):
    """Unit-test suite for `pptx.opc.package._Relationships` objects."""

    @pytest.mark.parametrize("rId, expected_value", (("rId1", True), ("rId2", False)))
    def it_knows_whether_it_contains_a_relationship_with_rId(
        self, _rels_prop_, rId, expected_value
    ):
        _rels_prop_.return_value = {"rId1": None}
        assert (rId in _Relationships(None)) is expected_value

    def it_has_dict_style_lookup_of_rel_by_rId(self, _rels_prop_, relationship_):
        _rels_prop_.return_value = {"rId17": relationship_}
        assert _Relationships(None)["rId17"] is relationship_

    def but_it_raises_KeyError_when_no_relationship_has_rId(self, _rels_prop_):
        _rels_prop_.return_value = {}
        with pytest.raises(KeyError) as e:
            _Relationships(None)["rId6"]
        assert str(e.value) == "\"no relationship with key 'rId6'\""

    def it_can_iterate_the_relationships_it_contains(self, request, _rels_prop_):
        rels_ = set(instance_mock(request, _Relationship) for n in range(5))
        _rels_prop_.return_value = {"rId%d" % (i + 1): r for i, r in enumerate(rels_)}
        relationships = _Relationships(None)

        for r in relationships:
            rels_.remove(r)

        assert len(rels_) == 0

    def it_has_a_len(self, _rels_prop_):
        _rels_prop_.return_value = {"a": 0, "b": 1}
        assert len(_Relationships(None)) == 2

    def it_can_add_a_relationship_to_a_target_part(
        self, part_, _get_matching_, _add_relationship_
    ):
        _get_matching_.return_value = None
        _add_relationship_.return_value = "rId7"
        relationships = _Relationships(None)

        rId = relationships.get_or_add(RT.IMAGE, part_)

        _get_matching_.assert_called_once_with(relationships, RT.IMAGE, part_)
        _add_relationship_.assert_called_once_with(relationships, RT.IMAGE, part_)
        assert rId == "rId7"

    def but_it_returns_an_existing_relationship_if_it_matches(
        self, part_, _get_matching_
    ):
        _get_matching_.return_value = "rId3"
        relationships = _Relationships(None)

        rId = relationships.get_or_add(RT.IMAGE, part_)

        _get_matching_.assert_called_once_with(relationships, RT.IMAGE, part_)
        assert rId == "rId3"

    def it_can_add_an_external_relationship_to_a_URI(
        self, _get_matching_, _add_relationship_
    ):
        _get_matching_.return_value = None
        _add_relationship_.return_value = "rId2"
        relationships = _Relationships(None)

        rId = relationships.get_or_add_ext_rel(RT.HYPERLINK, "http://url")

        _get_matching_.assert_called_once_with(
            relationships, RT.HYPERLINK, "http://url", is_external=True
        )
        _add_relationship_.assert_called_once_with(
            relationships, RT.HYPERLINK, "http://url", is_external=True
        )
        assert rId == "rId2"

    def but_it_returns_an_existing_external_relationship_if_it_matches(
        self, part_, _get_matching_
    ):
        _get_matching_.return_value = "rId10"
        relationships = _Relationships(None)

        rId = relationships.get_or_add_ext_rel(RT.HYPERLINK, "http://url")

        _get_matching_.assert_called_once_with(
            relationships, RT.HYPERLINK, "http://url", is_external=True
        )
        assert rId == "rId10"

    def it_can_load_from_the_xml_in_a_rels_part(self, request, _Relationship_, part_):
        rels_ = tuple(
            instance_mock(request, _Relationship, rId="rId%d" % (i + 1))
            for i in range(5)
        )
        _Relationship_.from_xml.side_effect = iter(rels_)
        parts = {"/ppt/slideLayouts/slideLayout1.xml": part_}
        xml_rels = parse_xml(snippet_bytes("rels-load-from-xml"))
        relationships = _Relationships(None)

        relationships.load_from_xml("/ppt/slides", xml_rels, parts)

        assert _Relationship_.from_xml.call_args_list == [
            call("/ppt/slides", xml_rels[0], parts),
            call("/ppt/slides", xml_rels[1], parts),
        ]
        assert relationships._rels == {"rId1": rels_[0], "rId2": rels_[1]}

    def it_can_find_a_part_with_reltype(
        self, _rels_by_reltype_prop_, relationship_, part_
    ):
        relationship_.target_part = part_
        _rels_by_reltype_prop_.return_value = collections.defaultdict(
            list, ((RT.SLIDE_LAYOUT, [relationship_]),)
        )
        relationships = _Relationships(None)

        assert relationships.part_with_reltype(RT.SLIDE_LAYOUT) is part_

    def but_it_raises_KeyError_when_there_is_no_such_part(self, _rels_by_reltype_prop_):
        _rels_by_reltype_prop_.return_value = collections.defaultdict(list)
        relationships = _Relationships(None)

        with pytest.raises(KeyError) as e:
            relationships.part_with_reltype(RT.SLIDE_LAYOUT)
        assert str(e.value) == (
            "\"no relationship of type 'http://schemas.openxmlformats.org/"
            "officeDocument/2006/relationships/slideLayout' in collection\""
        )

    def and_it_raises_ValueError_when_there_is_more_than_one_part_with_reltype(
        self, _rels_by_reltype_prop_, relationship_, part_
    ):
        relationship_.target_part = part_
        _rels_by_reltype_prop_.return_value = collections.defaultdict(
            list, ((RT.SLIDE_LAYOUT, [relationship_, relationship_]),)
        )
        relationships = _Relationships(None)

        with pytest.raises(ValueError) as e:
            relationships.part_with_reltype(RT.SLIDE_LAYOUT)
        assert str(e.value) == (
            "multiple relationships of type 'http://schemas.openxmlformats.org/"
            "officeDocument/2006/relationships/slideLayout' in collection"
        )

    def it_can_pop_a_relationship_to_remove_it_from_the_collection(
        self, _rels_prop_, relationship_
    ):
        _rels_prop_.return_value = {"rId22": relationship_}
        relationships = _Relationships(None)

        relationships.pop("rId22")

        assert relationships._rels == {}

    def it_can_serialize_itself_to_XML(self, request, _rels_prop_):
        _rels_prop_.return_value = {
            "rId1": instance_mock(
                request,
                _Relationship,
                rId="rId1",
                reltype=RT.SLIDE,
                target_ref="../slides/slide1.xml",
                is_external=False,
            ),
            "rId2": instance_mock(
                request,
                _Relationship,
                rId="rId2",
                reltype=RT.HYPERLINK,
                target_ref="http://url",
                is_external=True,
            ),
        }
        relationships = _Relationships(None)

        assert relationships.xml == snippet_bytes("relationships")

    def it_can_add_a_relationship_to_a_part_to_help(
        self,
        request,
        _next_rId_prop_,
        _Relationship_,
        relationship_,
        _rels_prop_,
        part_,
    ):
        _next_rId_prop_.return_value = "rId8"
        _Relationship_.return_value = relationship_
        _rels_prop_.return_value = {}
        relationships = _Relationships("/ppt")

        rId = relationships._add_relationship(RT.SLIDE, part_)

        _Relationship_.assert_called_once_with(
            "/ppt", "rId8", RT.SLIDE, target_mode=RTM.INTERNAL, target=part_
        )
        assert relationships._rels == {"rId8": relationship_}
        assert rId == "rId8"

    def and_it_can_add_an_external_relationship_to_help(
        self, request, _next_rId_prop_, _rels_prop_, _Relationship_, relationship_
    ):
        _next_rId_prop_.return_value = "rId9"
        _Relationship_.return_value = relationship_
        _rels_prop_.return_value = {}
        relationships = _Relationships("/ppt")

        rId = relationships._add_relationship(
            RT.HYPERLINK, "http://url", is_external=True
        )

        _Relationship_.assert_called_once_with(
            "/ppt", "rId9", RT.HYPERLINK, target_mode=RTM.EXTERNAL, target="http://url"
        )
        assert relationships._rels == {"rId9": relationship_}
        assert rId == "rId9"

    @pytest.mark.parametrize(
        "target_ref, is_external, expected_value",
        (
            ("http://url", True, "rId1"),
            ("part_1", False, "rId2"),
            ("http://foo", True, "rId3"),
            ("part_2", False, "rId4"),
            ("http://bar", True, None),
        ),
    )
    def it_can_get_a_matching_relationship_to_help(
        self,
        request,
        _rels_by_reltype_prop_,
        target_ref,
        is_external,
        expected_value,
    ):
        part_1, part_2 = (instance_mock(request, Part) for _ in range(2))
        _rels_by_reltype_prop_.return_value = {
            RT.SLIDE: [
                instance_mock(
                    request,
                    _Relationship,
                    rId=rId,
                    target_part=target_part,
                    target_ref=ref,
                    is_external=external,
                )
                for rId, target_part, ref, external in (
                    ("rId1", None, "http://url", True),
                    ("rId2", part_1, "/ppt/foo.bar", False),
                    ("rId3", None, "http://foo", True),
                    ("rId4", part_2, "/ppt/bar.foo", False),
                )
            ]
        }
        target = (
            target_ref if is_external else part_1 if target_ref == "part_1" else part_2
        )
        relationships = _Relationships(None)

        matching = relationships._get_matching(RT.SLIDE, target, is_external)

        assert matching == expected_value

    def but_it_returns_None_when_there_is_no_matching_relationship(
        self, _rels_by_reltype_prop_
    ):
        _rels_by_reltype_prop_.return_value = collections.defaultdict(list)
        relationships = _Relationships(None)

        assert relationships._get_matching(RT.HYPERLINK, "http://url", True) is None

    @pytest.mark.parametrize(
        "rIds, expected_value",
        (
            ((), "rId1"),
            (("rId1",), "rId2"),
            (("rId1", "rId2"), "rId3"),
            (("rId1", "rId4"), "rId3"),
            (("rId1", "rId4", "rId6"), "rId3"),
            (("rId1", "rId2", "rId6"), "rId4"),
        ),
    )
    def it_finds_the_next_rId_to_help(self, _rels_prop_, rIds, expected_value):
        _rels_prop_.return_value = {rId: None for rId in rIds}
        relationships = _Relationships(None)

        assert relationships._next_rId == expected_value

    def it_collects_relationships_by_reltype_to_help(self, request, _rels_prop_):
        rels = {
            "rId%d" % (i + 1): instance_mock(request, _Relationship, reltype=reltype)
            for i, reltype in enumerate((RT.SLIDE, RT.IMAGE, RT.SLIDE, RT.HYPERLINK))
        }
        _rels_prop_.return_value = rels
        relationships = _Relationships(None)

        rels_by_reltype = relationships._rels_by_reltype

        assert rels["rId1"] in rels_by_reltype[RT.SLIDE]
        assert rels["rId2"] in rels_by_reltype[RT.IMAGE]
        assert rels["rId3"] in rels_by_reltype[RT.SLIDE]
        assert rels["rId4"] in rels_by_reltype[RT.HYPERLINK]
        assert rels_by_reltype[RT.CHART] == []

    # fixture components -----------------------------------

    @pytest.fixture
    def _add_relationship_(self, request):
        return method_mock(request, _Relationships, "_add_relationship")

    @pytest.fixture
    def _get_matching_(self, request):
        return method_mock(request, _Relationships, "_get_matching")

    @pytest.fixture
    def _next_rId_prop_(self, request):
        return property_mock(request, _Relationships, "_next_rId")

    @pytest.fixture
    def part_(self, request):
        return instance_mock(request, Part)

    @pytest.fixture
    def _Relationship_(self, request):
        return class_mock(request, "pptx.opc.package._Relationship")

    @pytest.fixture
    def relationship_(self, request):
        return instance_mock(request, _Relationship)

    @pytest.fixture
    def _rels_by_reltype_prop_(self, request):
        return property_mock(request, _Relationships, "_rels_by_reltype")

    @pytest.fixture
    def _rels_prop_(self, request):
        return property_mock(request, _Relationships, "_rels")


class Describe_Relationship(object):
    """Unit-test suite for `pptx.opc.package._Relationship` objects."""

    def it_can_construct_from_xml(self, request, part_):
        _init_ = initializer_mock(request, _Relationship)
        rel_elm = instance_mock(
            request,
            CT_Relationship,
            rId="rId42",
            reltype=RT.SLIDE,
            targetMode=RTM.INTERNAL,
            target_ref="slides/slide7.xml",
        )
        parts = {"/ppt/slides/slide7.xml": part_}

        relationship = _Relationship.from_xml("/ppt", rel_elm, parts)

        _init_.assert_called_once_with(
            relationship,
            "/ppt",
            "rId42",
            RT.SLIDE,
            RTM.INTERNAL,
            part_,
        )
        assert isinstance(relationship, _Relationship)

    @pytest.mark.parametrize(
        "target_mode, expected_value",
        ((RTM.INTERNAL, False), (RTM.EXTERNAL, True), (None, False)),
    )
    def it_knows_whether_it_is_external(self, target_mode, expected_value):
        relationship = _Relationship(None, None, None, target_mode, None)
        assert relationship.is_external == expected_value

    def it_knows_its_relationship_type(self):
        relationship = _Relationship(None, None, RT.SLIDE, None, None)
        assert relationship.reltype == RT.SLIDE

    def it_knows_its_rId(self):
        relationship = _Relationship(None, "rId42", None, None, None)
        assert relationship.rId == "rId42"

    def it_provides_access_to_its_target_part(self, part_):
        relationship = _Relationship(None, None, None, RTM.INTERNAL, part_)
        assert relationship.target_part is part_

    def but_it_raises_ValueError_on_target_part_for_external_rel(self):
        relationship = _Relationship(None, None, None, RTM.EXTERNAL, None)
        with pytest.raises(ValueError) as e:
            relationship.target_part
        assert str(e.value) == (
            "`.target_part` property on _Relationship is undefined when "
            "target-mode is external"
        )

    def it_knows_its_target_partname(self, part_):
        part_.partname = PackURI("/ppt/slideLayouts/slideLayout4.xml")
        relationship = _Relationship(None, None, None, RTM.INTERNAL, part_)

        assert relationship.target_partname == "/ppt/slideLayouts/slideLayout4.xml"

    def but_it_raises_ValueError_on_target_partname_for_external_rel(self):
        relationship = _Relationship(None, None, None, RTM.EXTERNAL, None)

        with pytest.raises(ValueError) as e:
            relationship.target_partname

        assert str(e.value) == (
            "`.target_partname` property on _Relationship is undefined when "
            "target-mode is external"
        )

    def it_knows_the_target_uri_for_an_external_rel(self):
        relationship = _Relationship(None, None, None, RTM.EXTERNAL, "http://url")
        assert relationship.target_ref == "http://url"

    def and_it_knows_the_relative_partname_for_an_internal_rel(self, request):
        """Internal relationships have a relative reference for `.target_ref`.

        A relative reference looks like "../slideLayouts/slideLayout1.xml". This form
        is suitable for writing to a .rels file.
        """
        property_mock(
            request,
            _Relationship,
            "target_partname",
            return_value=PackURI("/ppt/media/image1.png"),
        )
        relationship = _Relationship("/ppt/slides", None, None, None, None)

        assert relationship.target_ref == "../media/image1.png"

    # --- fixture components -------------------------------

    @pytest.fixture
    def part_(self, request):
        return instance_mock(request, Part)



# encoding: utf-8

"""Unit-test suite for `pptx.parts.chart` module."""

import pytest

from pptx.chart.chart import Chart
from pptx.chart.data import ChartData
from pptx.enum.chart import XL_CHART_TYPE as XCT
from pptx.opc.constants import CONTENT_TYPE as CT, RELATIONSHIP_TYPE as RT
from pptx.opc.package import OpcPackage
from pptx.opc.packuri import PackURI
from pptx.oxml.chart.chart import CT_ChartSpace
from pptx.parts.chart import ChartPart, ChartWorkbook
from pptx.parts.embeddedpackage import EmbeddedXlsxPart

from ..unitutil.cxml import element, xml
from ..unitutil.mock import class_mock, instance_mock, method_mock, property_mock


class DescribeChartPart(object):
    """Unit-test suite for `pptx.parts.chart.ChartPart` objects."""

    def it_can_construct_from_chart_type_and_data(self, request):
        chart_data_ = instance_mock(request, ChartData, xlsx_blob=b"xlsx-blob")
        chart_data_.xml_bytes.return_value = b"chart-blob"
        package_ = instance_mock(request, OpcPackage)
        package_.next_partname.return_value = PackURI("/ppt/charts/chart42.xml")
        chart_part_ = instance_mock(request, ChartPart)
        # --- load() must have autospec turned off to work in Python 2.7 mock ---
        load_ = method_mock(
            request, ChartPart, "load", autospec=False, return_value=chart_part_
        )

        chart_part = ChartPart.new(XCT.RADAR, chart_data_, package_)

        package_.next_partname.assert_called_once_with("/ppt/charts/chart%d.xml")
        chart_data_.xml_bytes.assert_called_once_with(XCT.RADAR)
        load_.assert_called_once_with(
            "/ppt/charts/chart42.xml", CT.DML_CHART, package_, b"chart-blob"
        )
        chart_part_.chart_workbook.update_from_xlsx_blob.assert_called_once_with(
            b"xlsx-blob"
        )
        assert chart_part is chart_part_

    def it_provides_access_to_the_chart_object(self, request, chartSpace_):
        chart_ = instance_mock(request, Chart)
        Chart_ = class_mock(request, "pptx.parts.chart.Chart", return_value=chart_)
        chart_part = ChartPart(None, None, None, chartSpace_)

        chart = chart_part.chart

        Chart_.assert_called_once_with(chart_part._element, chart_part)
        assert chart is chart_

    def it_provides_access_to_the_chart_workbook(self, request, chartSpace_):
        chart_workbook_ = instance_mock(request, ChartWorkbook)
        ChartWorkbook_ = class_mock(
            request, "pptx.parts.chart.ChartWorkbook", return_value=chart_workbook_
        )
        chart_part = ChartPart(None, None, None, chartSpace_)

        chart_workbook = chart_part.chart_workbook

        ChartWorkbook_.assert_called_once_with(chartSpace_, chart_part)
        assert chart_workbook is chart_workbook_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def chartSpace_(self, request):
        return instance_mock(request, CT_ChartSpace)


class DescribeChartWorkbook(object):
    """Unit-test suite for `pptx.parts.chart.ChartWorkbook` objects."""

    def it_can_get_the_chart_xlsx_part(self, chart_part_, xlsx_part_):
        chart_part_.related_part.return_value = xlsx_part_
        chart_workbook = ChartWorkbook(
            element("c:chartSpace/c:externalData{r:id=rId42}"), chart_part_
        )

        xlsx_part = chart_workbook.xlsx_part

        chart_part_.related_part.assert_called_once_with("rId42")
        assert xlsx_part is xlsx_part_

    def but_it_returns_None_when_the_chart_has_no_xlsx_part(self):
        chart_workbook = ChartWorkbook(element("c:chartSpace"), None)
        assert chart_workbook.xlsx_part is None

    @pytest.mark.parametrize(
        "chartSpace_cxml, expected_cxml",
        (
            (
                "c:chartSpace{r:a=b}",
                "c:chartSpace{r:a=b}/c:externalData{r:id=rId" "42}/c:autoUpdate{val=0}",
            ),
            (
                "c:chartSpace/c:externalData{r:id=rId66}",
                "c:chartSpace/c:externalData{r:id=rId42}",
            ),
        ),
    )
    def it_can_change_the_chart_xlsx_part(
        self, chart_part_, xlsx_part_, chartSpace_cxml, expected_cxml
    ):
        chart_part_.relate_to.return_value = "rId42"
        chart_data = ChartWorkbook(element(chartSpace_cxml), chart_part_)

        chart_data.xlsx_part = xlsx_part_

        chart_part_.relate_to.assert_called_once_with(xlsx_part_, RT.PACKAGE)
        assert chart_data._chartSpace.xml == xml(expected_cxml)

    def it_adds_an_xlsx_part_on_update_if_needed(
        self, request, chart_part_, package_, xlsx_part_, xlsx_part_prop_
    ):
        EmbeddedXlsxPart_ = class_mock(request, "pptx.parts.chart.EmbeddedXlsxPart")
        EmbeddedXlsxPart_.new.return_value = xlsx_part_
        chart_part_.package = package_
        xlsx_part_prop_.return_value = None
        chart_data = ChartWorkbook(element("c:chartSpace"), chart_part_)

        chart_data.update_from_xlsx_blob(b"xlsx-blob")

        EmbeddedXlsxPart_.new.assert_called_once_with(b"xlsx-blob", package_)
        xlsx_part_prop_.assert_called_with(xlsx_part_)

    def but_it_replaces_the_xlsx_blob_when_the_part_exists(
        self, xlsx_part_prop_, xlsx_part_
    ):
        xlsx_part_prop_.return_value = xlsx_part_
        chart_data = ChartWorkbook(None, None)
        chart_data.update_from_xlsx_blob(b"xlsx-blob")

        assert chart_data.xlsx_part.blob == b"xlsx-blob"

    # fixture components ---------------------------------------------

    @pytest.fixture
    def chart_part_(self, request, package_, xlsx_part_):
        return instance_mock(request, ChartPart)

    @pytest.fixture
    def package_(self, request):
        return instance_mock(request, OpcPackage)

    @pytest.fixture
    def xlsx_part_(self, request):
        return instance_mock(request, EmbeddedXlsxPart)

    @pytest.fixture
    def xlsx_part_prop_(self, request):
        return property_mock(request, ChartWorkbook, "xlsx_part")



# encoding: utf-8

"""Unit-test suite for `pptx.opc.serialized` module."""

import hashlib
import pytest
import zipfile

from pptx.compat import BytesIO
from pptx.exceptions import PackageNotFoundError
from pptx.opc.constants import CONTENT_TYPE as CT
from pptx.opc.package import Part, _Relationships
from pptx.opc.packuri import CONTENT_TYPES_URI, PackURI
from pptx.opc.serialized import (
    PackageReader,
    PackageWriter,
    _ContentTypesItem,
    _DirPkgReader,
    _PhysPkgReader,
    _PhysPkgWriter,
    _ZipPkgReader,
    _ZipPkgWriter,
)

from ..unitutil.file import absjoin, snippet_text, test_file_dir
from ..unitutil.mock import (
    ANY,
    call,
    class_mock,
    function_mock,
    initializer_mock,
    instance_mock,
    method_mock,
    property_mock,
)


test_pptx_path = absjoin(test_file_dir, "test.pptx")
dir_pkg_path = absjoin(test_file_dir, "expanded_pptx")
zip_pkg_path = test_pptx_path


class DescribePackageReader(object):
    """Unit-test suite for `pptx.opc.serialized.PackageReader` objects."""

    def it_knows_whether_it_contains_a_partname(self, _blob_reader_prop_):
        _blob_reader_prop_.return_value = set(("/ppt", "/docProps"))
        package_reader = PackageReader(None)

        assert "/ppt" in package_reader
        assert "/xyz" not in package_reader

    def it_can_get_a_blob_by_partname(self, _blob_reader_prop_):
        _blob_reader_prop_.return_value = {"/ppt/slides/slide1.xml": b"blob"}
        package_reader = PackageReader(None)

        assert package_reader["/ppt/slides/slide1.xml"] == b"blob"

    def it_can_get_the_rels_xml_for_a_partname(self, _blob_reader_prop_):
        _blob_reader_prop_.return_value = {"/ppt/_rels/presentation.xml.rels": b"blob"}
        package_reader = PackageReader(None)

        assert package_reader.rels_xml_for(PackURI("/ppt/presentation.xml")) == b"blob"

    def but_it_returns_None_when_the_part_has_no_rels(self, _blob_reader_prop_):
        _blob_reader_prop_.return_value = {"/ppt/_rels/presentation.xml.rels": b"blob"}
        package_reader = PackageReader(None)

        assert package_reader.rels_xml_for(PackURI("/ppt/slides.slide1.xml")) is None

    def it_constructs_its_blob_reader_to_help(self, request):
        phys_pkg_reader_ = instance_mock(request, _PhysPkgReader)
        _PhysPkgReader_ = class_mock(request, "pptx.opc.serialized._PhysPkgReader")
        _PhysPkgReader_.factory.return_value = phys_pkg_reader_
        package_reader = PackageReader("prs.pptx")

        blob_reader = package_reader._blob_reader

        _PhysPkgReader_.factory.assert_called_once_with("prs.pptx")
        assert blob_reader is phys_pkg_reader_

    # fixture components -----------------------------------

    @pytest.fixture
    def _blob_reader_prop_(self, request):
        return property_mock(request, PackageReader, "_blob_reader")


class DescribePackageWriter(object):
    """Unit-test suite for `pptx.opc.serialized.PackageWriter` objects."""

    def it_provides_a_write_interface_classmethod(self, request, relationships_):
        _init_ = initializer_mock(request, PackageWriter)
        _write_ = method_mock(request, PackageWriter, "_write")

        PackageWriter.write("prs.pptx", relationships_, ("part_1", "part_2"))

        _init_.assert_called_once_with(
            ANY, "prs.pptx", relationships_, ("part_1", "part_2")
        )
        _write_.assert_called_once_with(ANY)

    def it_can_write_a_package(self, request, phys_writer_):
        _PhysPkgWriter_ = class_mock(request, "pptx.opc.serialized._PhysPkgWriter")
        phys_writer_.__enter__.return_value = phys_writer_
        _PhysPkgWriter_.factory.return_value = phys_writer_
        _write_content_types_stream_ = method_mock(
            request, PackageWriter, "_write_content_types_stream"
        )
        _write_pkg_rels_ = method_mock(request, PackageWriter, "_write_pkg_rels")
        _write_parts_ = method_mock(request, PackageWriter, "_write_parts")
        package_writer = PackageWriter("prs.pptx", None, None)

        package_writer._write()

        _PhysPkgWriter_.factory.assert_called_once_with("prs.pptx")
        _write_content_types_stream_.assert_called_once_with(
            package_writer, phys_writer_
        )
        _write_pkg_rels_.assert_called_once_with(package_writer, phys_writer_)
        _write_parts_.assert_called_once_with(package_writer, phys_writer_)

    def it_can_write_a_content_types_stream(self, request, phys_writer_):
        _ContentTypesItem_ = class_mock(
            request, "pptx.opc.serialized._ContentTypesItem"
        )
        _ContentTypesItem_.xml_for.return_value = "part_xml"
        serialize_part_xml_ = function_mock(
            request, "pptx.opc.serialized.serialize_part_xml", return_value=b"xml"
        )
        package_writer = PackageWriter(None, None, ("part_1", "part_2"))

        package_writer._write_content_types_stream(phys_writer_)

        _ContentTypesItem_.xml_for.assert_called_once_with(("part_1", "part_2"))
        serialize_part_xml_.assert_called_once_with("part_xml")
        phys_writer_.write.assert_called_once_with(CONTENT_TYPES_URI, b"xml")

    def it_can_write_a_sequence_of_parts(self, request, phys_writer_):
        parts_ = (
            instance_mock(
                request,
                Part,
                partname=PackURI("/ppt/%s.xml" % x),
                blob="blob_%s" % x,
                rels=instance_mock(request, _Relationships, xml="rels_xml_%s" % x),
            )
            for x in ("a", "b", "c")
        )
        package_writer = PackageWriter(None, None, parts_)

        package_writer._write_parts(phys_writer_)

        assert phys_writer_.write.call_args_list == [
            call("/ppt/a.xml", "blob_a"),
            call("/ppt/_rels/a.xml.rels", "rels_xml_a"),
            call("/ppt/b.xml", "blob_b"),
            call("/ppt/_rels/b.xml.rels", "rels_xml_b"),
            call("/ppt/c.xml", "blob_c"),
            call("/ppt/_rels/c.xml.rels", "rels_xml_c"),
        ]

    def it_can_write_a_pkg_rels_item(self, request, phys_writer_, relationships_):
        relationships_.xml = b"pkg-rels-xml"
        package_writer = PackageWriter(None, relationships_, None)

        package_writer._write_pkg_rels(phys_writer_)

        phys_writer_.write.assert_called_once_with("/_rels/.rels", b"pkg-rels-xml")

    # fixture components -----------------------------------

    @pytest.fixture
    def phys_writer_(self, request):
        return instance_mock(request, _ZipPkgWriter)

    @pytest.fixture
    def relationships_(self, request):
        return instance_mock(request, _Relationships)


class Describe_PhysPkgReader(object):
    """Unit-test suite for `pptx.opc.serialized._PhysPkgReader` objects."""

    def it_constructs_ZipPkgReader_when_pkg_is_file_like(
        self, _ZipPkgReader_, zip_pkg_reader_
    ):
        _ZipPkgReader_.return_value = zip_pkg_reader_
        file_like_pkg = BytesIO(b"pkg-bytes")

        phys_reader = _PhysPkgReader.factory(file_like_pkg)

        _ZipPkgReader_.assert_called_once_with(file_like_pkg)
        assert phys_reader is zip_pkg_reader_

    def and_it_constructs_DirPkgReader_when_pkg_is_a_dir(self, request):
        dir_pkg_reader_ = instance_mock(request, _DirPkgReader)
        _DirPkgReader_ = class_mock(
            request, "pptx.opc.serialized._DirPkgReader", return_value=dir_pkg_reader_
        )

        phys_reader = _PhysPkgReader.factory(dir_pkg_path)

        _DirPkgReader_.assert_called_once_with(dir_pkg_path)
        assert phys_reader is dir_pkg_reader_

    def and_it_constructs_ZipPkgReader_when_pkg_is_a_zip_file_path(
        self, _ZipPkgReader_, zip_pkg_reader_
    ):
        _ZipPkgReader_.return_value = zip_pkg_reader_
        pkg_file_path = test_pptx_path

        phys_reader = _PhysPkgReader.factory(pkg_file_path)

        _ZipPkgReader_.assert_called_once_with(pkg_file_path)
        assert phys_reader is zip_pkg_reader_

    def but_it_raises_when_pkg_path_is_not_a_package(self):
        with pytest.raises(PackageNotFoundError) as e:
            _PhysPkgReader.factory("foobar")
        assert str(e.value) == "Package not found at 'foobar'"

    # --- fixture components -------------------------------

    @pytest.fixture
    def zip_pkg_reader_(self, request):
        return instance_mock(request, _ZipPkgReader)

    @pytest.fixture
    def _ZipPkgReader_(self, request):
        return class_mock(request, "pptx.opc.serialized._ZipPkgReader")


class Describe_DirPkgReader(object):
    """Unit-test suite for `pptx.opc.serialized._DirPkgReader` objects."""

    def it_can_retrieve_the_blob_for_a_pack_uri(self):
        blob = _DirPkgReader(dir_pkg_path)[PackURI("/ppt/presentation.xml")]

        sha1 = hashlib.sha1(blob).hexdigest()
        assert sha1 == "51b78f4dabc0af2419d4e044ab73028c4bef53aa"

    def but_it_raises_KeyError_when_requested_member_is_not_present(self):
        with pytest.raises(KeyError) as e:
            _DirPkgReader(dir_pkg_path)[PackURI("/ppt/foobar.xml")]
        assert str(e.value) == "\"no member '/ppt/foobar.xml' in package\""


class Describe_ZipPkgReader(object):
    """Unit-test suite for `pptx.opc.serialized._ZipPkgReader` objects."""

    def it_knows_whether_it_contains_a_partname(self, zip_pkg_reader):
        assert PackURI("/ppt/presentation.xml") in zip_pkg_reader
        assert PackURI("/ppt/foobar.xml") not in zip_pkg_reader

    def it_can_get_a_blob_by_partname(self, zip_pkg_reader):
        blob = zip_pkg_reader[PackURI("/ppt/presentation.xml")]
        assert hashlib.sha1(blob).hexdigest() == (
            "efa7bee0ac72464903a67a6744c1169035d52a54"
        )

    def but_it_raises_KeyError_when_requested_member_is_not_present(
        self, zip_pkg_reader
    ):
        with pytest.raises(KeyError) as e:
            zip_pkg_reader[PackURI("/ppt/foobar.xml")]
        assert str(e.value) == "\"no member '/ppt/foobar.xml' in package\""

    def it_loads_the_package_blobs_on_first_access_to_help(self, zip_pkg_reader):
        blobs = zip_pkg_reader._blobs
        assert len(blobs) == 38
        assert "/ppt/presentation.xml" in blobs
        assert "/ppt/_rels/presentation.xml.rels" in blobs

    # --- fixture components -------------------------------

    @pytest.fixture(scope="class")
    def zip_pkg_reader(self, request):
        return _ZipPkgReader(zip_pkg_path)


class Describe_PhysPkgWriter(object):
    """Unit-test suite for `pptx.opc.serialized._PhysPkgWriter` objects."""

    def it_constructs_ZipPkgWriter_unconditionally(self, request):
        zip_pkg_writer_ = instance_mock(request, _ZipPkgWriter)
        _ZipPkgWriter_ = class_mock(
            request, "pptx.opc.serialized._ZipPkgWriter", return_value=zip_pkg_writer_
        )

        phys_writer = _PhysPkgWriter.factory("prs.pptx")

        _ZipPkgWriter_.assert_called_once_with("prs.pptx")
        assert phys_writer is zip_pkg_writer_


class Describe_ZipPkgWriter(object):
    """Unit-test suite for `pptx.opc.serialized._ZipPkgWriter` objects."""

    def it_has_an__enter__method_for_context_management(self):
        pkg_writer = _ZipPkgWriter(None)
        assert pkg_writer.__enter__() is pkg_writer

    def and_it_closes_the_zip_archive_on_context__exit__(self, _zipf_prop_):
        _ZipPkgWriter(None).__exit__(None, None, None)
        _zipf_prop_.return_value.close.assert_called_once_with()

    def it_can_write_a_blob(self, _zipf_prop_):
        """Integrates with zipfile.ZipFile."""
        pack_uri = PackURI("/part/name.xml")
        _zipf_prop_.return_value = zipf = zipfile.ZipFile(BytesIO(), "w")
        pkg_writer = _ZipPkgWriter(None)

        pkg_writer.write(pack_uri, b"blob")

        members = {PackURI("/%s" % name): zipf.read(name) for name in zipf.namelist()}
        assert len(members) == 1
        assert members[pack_uri] == b"blob"

    def it_provides_access_to_the_open_zip_file_to_help(self, request):
        ZipFile_ = class_mock(request, "pptx.opc.serialized.zipfile.ZipFile")
        pkg_writer = _ZipPkgWriter("prs.pptx")

        zipf = pkg_writer._zipf

        ZipFile_.assert_called_once_with(
            "prs.pptx", "w", compression=zipfile.ZIP_DEFLATED
        )
        assert zipf is ZipFile_.return_value

    # fixtures ---------------------------------------------

    @pytest.fixture
    def _zipf_prop_(self, request):
        return property_mock(request, _ZipPkgWriter, "_zipf")


class Describe_ContentTypesItem(object):
    """Unit-test suite for `pptx.opc.serialized._ContentTypesItem` objects."""

    def it_provides_an_interface_classmethod(self, request):
        _init_ = initializer_mock(request, _ContentTypesItem)
        property_mock(request, _ContentTypesItem, "_xml", return_value=b"xml")

        xml = _ContentTypesItem.xml_for(("part", "zuh"))

        _init_.assert_called_once_with(ANY, ("part", "zuh"))
        assert xml == b"xml"

    def it_can_compose_content_types_xml(self, request):
        defaults = {"png": CT.PNG, "xml": CT.XML, "rels": CT.OPC_RELATIONSHIPS}
        overrides = {
            "/docProps/core.xml": "app/vnd.core",
            "/ppt/slides/slide1.xml": "app/vnd.ct_sld",
            "/zebra/foo.bar": "app/vnd.foobar",
        }
        property_mock(
            request,
            _ContentTypesItem,
            "_defaults_and_overrides",
            return_value=(defaults, overrides),
        )

        content_types = _ContentTypesItem(None)._xml

        assert content_types.xml == snippet_text("content-types-xml").strip()

    def it_computes_defaults_and_overrides_to_help(self, request):
        parts = (
            instance_mock(
                request, Part, partname=PackURI(partname), content_type=content_type
            )
            for partname, content_type in (
                ("/media/image1.png", CT.PNG),
                ("/ppt/slides/slide1.xml", CT.PML_SLIDE),
                ("/foo/bar.xml", CT.XML),
                ("/docProps/core.xml", CT.OPC_CORE_PROPERTIES),
            )
        )
        content_types = _ContentTypesItem(parts)

        defaults, overrides = content_types._defaults_and_overrides

        assert defaults == {"png": CT.PNG, "rels": CT.OPC_RELATIONSHIPS, "xml": CT.XML}
        assert overrides == {
            "/ppt/slides/slide1.xml": CT.PML_SLIDE,
            "/docProps/core.xml": CT.OPC_CORE_PROPERTIES,
        }



# encoding: utf-8

"""API for reading/writing serialized Open Packaging Convention (OPC) package."""

import os
import zipfile

from pptx.compat import Container, is_string
from pptx.exceptions import PackageNotFoundError
from pptx.opc.constants import CONTENT_TYPE as CT
from pptx.opc.oxml import CT_Types, serialize_part_xml
from pptx.opc.packuri import CONTENT_TYPES_URI, PACKAGE_URI, PackURI
from pptx.opc.shared import CaseInsensitiveDict
from pptx.opc.spec import default_content_types
from pptx.util import lazyproperty


class PackageReader(Container):
    """Provides access to package-parts of an OPC package with dict semantics.

    The package may be in zip-format (a .pptx file) or expanded into a directory
    structure, perhaps by unzipping a .pptx file.
    """

    def __init__(self, pkg_file):
        self._pkg_file = pkg_file

    def __contains__(self, pack_uri):
        """Return True when part identified by `pack_uri` is present in package."""
        return pack_uri in self._blob_reader

    def __getitem__(self, pack_uri):
        """Return bytes for part corresponding to `pack_uri`."""
        return self._blob_reader[pack_uri]

    def rels_xml_for(self, partname):
        """Return optional rels item XML for `partname`.

        Returns `None` if no rels item is present for `partname`. `partname` is a
        |PackURI| instance.
        """
        blob_reader, uri = self._blob_reader, partname.rels_uri
        return blob_reader[uri] if uri in blob_reader else None

    @lazyproperty
    def _blob_reader(self):
        """|_PhysPkgReader| subtype providing read access to the package file."""
        return _PhysPkgReader.factory(self._pkg_file)


class PackageWriter(object):
    """Writes a zip-format OPC package to `pkg_file`.

    `pkg_file` can be either a path to a zip file (a string) or a file-like object.
    `pkg_rels` is the |_Relationships| object containing relationships for the package.
    `parts` is a sequence of |Part| subtype instance to be written to the package.

    Its single API classmethod is :meth:`write`. This class is not intended to be
    instantiated.
    """

    def __init__(self, pkg_file, pkg_rels, parts):
        self._pkg_file = pkg_file
        self._pkg_rels = pkg_rels
        self._parts = parts

    @classmethod
    def write(cls, pkg_file, pkg_rels, parts):
        """Write a physical package (.pptx file) to `pkg_file`.

        The serialized package contains `pkg_rels` and `parts`, a content-types stream
        based on the content type of each part, and a .rels file for each part that has
        relationships.
        """
        cls(pkg_file, pkg_rels, parts)._write()

    def _write(self):
        """Write physical package (.pptx file)."""
        with _PhysPkgWriter.factory(self._pkg_file) as phys_writer:
            self._write_content_types_stream(phys_writer)
            self._write_pkg_rels(phys_writer)
            self._write_parts(phys_writer)

    def _write_content_types_stream(self, phys_writer):
        """Write `[Content_Types].xml` part to the physical package.

        This part must contain an appropriate content type lookup target for each part
        in the package.
        """
        phys_writer.write(
            CONTENT_TYPES_URI,
            serialize_part_xml(_ContentTypesItem.xml_for(self._parts)),
        )

    def _write_parts(self, phys_writer):
        """Write blob of each part in `parts` to the package.

        A rels item for each part is also written when the part has relationships.
        """
        for part in self._parts:
            phys_writer.write(part.partname, part.blob)
            if part._rels:
                phys_writer.write(part.partname.rels_uri, part.rels.xml)

    def _write_pkg_rels(self, phys_writer):
        """Write the XML rels item for *pkg_rels* ('/_rels/.rels') to the package."""
        phys_writer.write(PACKAGE_URI.rels_uri, self._pkg_rels.xml)


class _PhysPkgReader(Container):
    """Base class for physical package reader objects."""

    def __contains__(self, item):
        """Must be implemented by each subclass."""
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.__contains__()`" % type(self).__name__
        )

    @classmethod
    def factory(cls, pkg_file):
        """Return |_PhysPkgReader| subtype instance appropriage for `pkg_file`."""
        # --- for pkg_file other than str, assume it's a stream and pass it to Zip
        # --- reader to sort out
        if not is_string(pkg_file):
            return _ZipPkgReader(pkg_file)

        # --- otherwise we treat `pkg_file` as a path ---
        if os.path.isdir(pkg_file):
            return _DirPkgReader(pkg_file)

        if zipfile.is_zipfile(pkg_file):
            return _ZipPkgReader(pkg_file)

        raise PackageNotFoundError("Package not found at '%s'" % pkg_file)


class _DirPkgReader(_PhysPkgReader):
    """Implements |PhysPkgReader| interface for OPC package extracted into directory.

    `path` is the path to a directory containing an expanded package.
    """

    def __init__(self, path):
        self._path = os.path.abspath(path)

    def __getitem__(self, pack_uri):
        """Return bytes of file corresponding to `pack_uri` in package directory."""
        path = os.path.join(self._path, pack_uri.membername)
        try:
            with open(path, "rb") as f:
                return f.read()
        except IOError:
            raise KeyError("no member '%s' in package" % pack_uri)


class _ZipPkgReader(_PhysPkgReader):
    """Implements |PhysPkgReader| interface for a zip-file OPC package."""

    def __init__(self, pkg_file):
        self._pkg_file = pkg_file

    def __contains__(self, pack_uri):
        """Return True when part identified by `pack_uri` is present in zip archive."""
        return pack_uri in self._blobs

    def __getitem__(self, pack_uri):
        """Return bytes for part corresponding to `pack_uri`.

        Raises |KeyError| if no matching member is present in zip archive.
        """
        if pack_uri not in self._blobs:
            raise KeyError("no member '%s' in package" % pack_uri)
        return self._blobs[pack_uri]

    @lazyproperty
    def _blobs(self):
        """dict mapping partname to package part binaries."""
        with zipfile.ZipFile(self._pkg_file, "r") as z:
            return {PackURI("/%s" % name): z.read(name) for name in z.namelist()}


class _PhysPkgWriter(object):
    """Base class for physical package writer objects."""

    @classmethod
    def factory(cls, pkg_file):
        """Return |_PhysPkgWriter| subtype instance appropriage for `pkg_file`.

        Currently the only subtype is `_ZipPkgWriter`, but a `_DirPkgWriter` could be
        implemented or even a `_StreamPkgWriter`.
        """
        return _ZipPkgWriter(pkg_file)


class _ZipPkgWriter(_PhysPkgWriter):
    """Implements |PhysPkgWriter| interface for a zip-file (.pptx file) OPC package."""

    def __init__(self, pkg_file):
        self._pkg_file = pkg_file

    def __enter__(self):
        """Enable use as a context-manager. Opening zip for writing happens here."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Close the zip archive on exit from context.

        Closing flushes any pending physical writes and releasing any resources it's
        using.
        """
        self._zipf.close()

    def write(self, pack_uri, blob):
        """Write `blob` to zip package with membername corresponding to `pack_uri`."""
        self._zipf.writestr(pack_uri.membername, blob)

    @lazyproperty
    def _zipf(self):
        """`ZipFile` instance open for writing."""
        return zipfile.ZipFile(self._pkg_file, "w", compression=zipfile.ZIP_DEFLATED)


class _ContentTypesItem(object):
    """Composes content-types "part" ([Content_Types].xml) for a collection of parts."""

    def __init__(self, parts):
        self._parts = parts

    @classmethod
    def xml_for(cls, parts):
        """Return content-types XML mapping each part in `parts` to a content-type.

        The resulting XML is suitable for storage as `[Content_Types].xml` in an OPC
        package.
        """
        return cls(parts)._xml

    @lazyproperty
    def _xml(self):
        """lxml.etree._Element containing the content-types item.

        This XML object is suitable for serialization to the `[Content_Types].xml` item
        for an OPC package. Although the sequence of elements is not strictly
        significant, as an aid to testing and readability Default elements are sorted by
        extension and Override elements are sorted by partname.
        """
        defaults, overrides = self._defaults_and_overrides
        _types_elm = CT_Types.new()

        for ext, content_type in sorted(defaults.items()):
            _types_elm.add_default(ext, content_type)
        for partname, content_type in sorted(overrides.items()):
            _types_elm.add_override(partname, content_type)

        return _types_elm

    @lazyproperty
    def _defaults_and_overrides(self):
        """pair of dict (defaults, overrides) accounting for all parts.

        `defaults` is {ext: content_type} and overrides is {partname: content_type}.
        """
        defaults = CaseInsensitiveDict(rels=CT.OPC_RELATIONSHIPS, xml=CT.XML)
        overrides = dict()

        for part in self._parts:
            partname, content_type = part.partname, part.content_type
            ext = partname.ext
            if (ext.lower(), content_type) in default_content_types:
                defaults[ext] = content_type
            else:
                overrides[partname] = content_type

        return defaults, overrides



# encoding: utf-8

"""Unit-test suite for `pptx.oxml.slide` module."""

from pptx.oxml.slide import CT_NotesMaster, CT_NotesSlide

from ..unitutil.file import snippet_text


class DescribeCT_NotesMaster(object):
    """Unit-test suite for `pptx.oxml.slide.CT_NotesMaster` objects."""

    def it_can_create_a_default_notesMaster_element(self):
        notesMaster = CT_NotesMaster.new_default()
        assert notesMaster.xml == snippet_text("default-notesMaster")


class DescribeCT_NotesSlide(object):
    """Unit-test suite for `pptx.oxml.slide.CT_NotesSlide` objects."""

    def it_can_create_a_new_notes_element(self):
        notes = CT_NotesSlide.new()
        assert notes.xml == snippet_text("default-notes")



# encoding: utf-8

"""
XML test data builders for pptx.oxml.text unit tests
"""

from __future__ import absolute_import, print_function

from ...unitdata import BaseBuilder


class CT_Hyperlink(BaseBuilder):
    __tag__ = "a:hlinkClick"
    __nspfxs__ = ("a", "r")
    __attrs__ = (
        "r:id",
        "invalidUrl",
        "action",
        "tgtFrame",
        "tooltip",
        "history",
        "highlightClick",
        "endSnd",
    )

    def with_rId(self, rId):
        self._set_xmlattr("r:id", rId)
        return self


class CT_RegularTextRunBuilder(BaseBuilder):
    __tag__ = "a:r"
    __nspfxs__ = ("a",)
    __attrs__ = ()


class CT_TextCharacterPropertiesBuilder(BaseBuilder):
    """
    Test data builder for CT_TextCharacterProperties XML element that appears
    as <a:endParaRPr> child of <a:p> and <a:rPr> child of <a:r>.
    """

    __nspfxs__ = ("a",)
    __attrs__ = ("b", "i", "sz", "u")

    def __init__(self, tag):
        self.__tag__ = tag
        super(CT_TextCharacterPropertiesBuilder, self).__init__()


class CT_TextParagraphBuilder(BaseBuilder):
    """
    Test data builder for CT_TextParagraph (<a:p>) XML element that appears
    as a child of <p:txBody>.
    """

    __tag__ = "a:p"
    __nspfxs__ = ("a",)
    __attrs__ = ()


class XsdString(BaseBuilder):
    __attrs__ = ()

    def __init__(self, tag, nspfxs):
        self.__tag__ = tag
        self.__nspfxs__ = nspfxs
        super(XsdString, self).__init__()


def a_p():
    """Return a CT_TextParagraphBuilder instance"""
    return CT_TextParagraphBuilder()


def a_t():
    return XsdString("a:t", ("a",))


def an_hlinkClick():
    return CT_Hyperlink()


def an_r():
    return CT_RegularTextRunBuilder()


def an_rPr():
    return CT_TextCharacterPropertiesBuilder("a:rPr")



# encoding: utf-8

"""OPC-local oxml module to handle OPC-local concerns like relationship parsing."""

from lxml import etree

from .constants import NAMESPACE as NS, RELATIONSHIP_TARGET_MODE as RTM
from ..oxml import parse_xml, register_element_cls
from ..oxml.simpletypes import (
    ST_ContentType,
    ST_Extension,
    ST_TargetMode,
    XsdAnyUri,
    XsdId,
)
from ..oxml.xmlchemy import (
    BaseOxmlElement,
    OptionalAttribute,
    RequiredAttribute,
    ZeroOrMore,
)


nsmap = {
    "ct": NS.OPC_CONTENT_TYPES,
    "pr": NS.OPC_RELATIONSHIPS,
    "r": NS.OFC_RELATIONSHIPS,
}


def oxml_tostring(elm, encoding=None, pretty_print=False, standalone=None):
    return etree.tostring(
        elm, encoding=encoding, pretty_print=pretty_print, standalone=standalone
    )


def serialize_part_xml(part_elm):
    xml = etree.tostring(part_elm, encoding="UTF-8", standalone=True)
    return xml


class CT_Default(BaseOxmlElement):
    """
    ``<Default>`` element, specifying the default content type to be applied
    to a part with the specified extension.
    """

    extension = RequiredAttribute("Extension", ST_Extension)
    contentType = RequiredAttribute("ContentType", ST_ContentType)


class CT_Override(BaseOxmlElement):
    """
    ``<Override>`` element, specifying the content type to be applied for a
    part with the specified partname.
    """

    partName = RequiredAttribute("PartName", XsdAnyUri)
    contentType = RequiredAttribute("ContentType", ST_ContentType)


class CT_Relationship(BaseOxmlElement):
    """
    ``<Relationship>`` element, representing a single relationship from a
    source to a target part.
    """

    rId = RequiredAttribute("Id", XsdId)
    reltype = RequiredAttribute("Type", XsdAnyUri)
    target_ref = RequiredAttribute("Target", XsdAnyUri)
    targetMode = OptionalAttribute("TargetMode", ST_TargetMode, default=RTM.INTERNAL)

    @classmethod
    def new(cls, rId, reltype, target, target_mode=RTM.INTERNAL):
        """
        Return a new ``<Relationship>`` element.
        """
        xml = '<Relationship xmlns="%s"/>' % nsmap["pr"]
        relationship = parse_xml(xml)
        relationship.rId = rId
        relationship.reltype = reltype
        relationship.target_ref = target
        relationship.targetMode = target_mode
        return relationship


class CT_Relationships(BaseOxmlElement):
    """`<Relationships>` element, the root element in a .rels file."""

    relationship = ZeroOrMore("pr:Relationship")

    def add_rel(self, rId, reltype, target, is_external=False):
        """
        Add a child ``<Relationship>`` element with attributes set according
        to parameter values.
        """
        target_mode = RTM.EXTERNAL if is_external else RTM.INTERNAL
        relationship = CT_Relationship.new(rId, reltype, target, target_mode)
        self._insert_relationship(relationship)

    @classmethod
    def new(cls):
        """Return a new ``<Relationships>`` element."""
        return parse_xml('<Relationships xmlns="%s"/>' % nsmap["pr"])

    @property
    def xml(self):
        """
        Return XML string for this element, suitable for saving in a .rels
        stream, not pretty printed and with an XML declaration at the top.
        """
        return oxml_tostring(self, encoding="UTF-8", standalone=True)


class CT_Types(BaseOxmlElement):
    """
    ``<Types>`` element, the container element for Default and Override
    elements in [Content_Types].xml.
    """

    default = ZeroOrMore("ct:Default")
    override = ZeroOrMore("ct:Override")

    def add_default(self, ext, content_type):
        """
        Add a child ``<Default>`` element with attributes set to parameter
        values.
        """
        return self._add_default(extension=ext, contentType=content_type)

    def add_override(self, partname, content_type):
        """
        Add a child ``<Override>`` element with attributes set to parameter
        values.
        """
        return self._add_override(partName=partname, contentType=content_type)

    @classmethod
    def new(cls):
        """
        Return a new ``<Types>`` element.
        """
        xml = '<Types xmlns="%s"/>' % nsmap["ct"]
        types = parse_xml(xml)
        return types


register_element_cls("ct:Default", CT_Default)
register_element_cls("ct:Override", CT_Override)
register_element_cls("ct:Types", CT_Types)

register_element_cls("pr:Relationship", CT_Relationship)
register_element_cls("pr:Relationships", CT_Relationships)



# encoding: utf-8

"""Fundamental Open Packaging Convention (OPC) objects.

The :mod:`pptx.packaging` module coheres around the concerns of reading and writing
presentations to and from a .pptx file.
"""

import collections

from pptx.compat import is_string, Mapping
from pptx.opc.constants import RELATIONSHIP_TARGET_MODE as RTM, RELATIONSHIP_TYPE as RT
from pptx.opc.oxml import CT_Relationships, serialize_part_xml
from pptx.opc.packuri import CONTENT_TYPES_URI, PACKAGE_URI, PackURI
from pptx.opc.serialized import PackageReader, PackageWriter
from pptx.opc.shared import CaseInsensitiveDict
from pptx.oxml import parse_xml
from pptx.util import lazyproperty


class _RelatableMixin(object):
    """Provide relationship methods required by both the package and each part."""

    def part_related_by(self, reltype):
        """Return (single) part having relationship to this package of `reltype`.

        Raises |KeyError| if no such relationship is found and |ValueError| if more than
        one such relationship is found.
        """
        return self._rels.part_with_reltype(reltype)

    def relate_to(self, target, reltype, is_external=False):
        """Return rId key of relationship of `reltype` to `target`.

        If such a relationship already exists, its rId is returned. Otherwise the
        relationship is added and its new rId returned.
        """
        return (
            self._rels.get_or_add_ext_rel(reltype, target)
            if is_external
            else self._rels.get_or_add(reltype, target)
        )

    def related_part(self, rId):
        """Return related |Part| subtype identified by `rId`."""
        return self._rels[rId].target_part

    def target_ref(self, rId):
        """Return URL contained in target ref of relationship identified by `rId`."""
        return self._rels[rId].target_ref

    @lazyproperty
    def _rels(self):
        """|Relationships| object containing relationships from this part to others."""
        raise NotImplementedError(  # pragma: no cover
            "`%s` must implement `.rels`" % type(self).__name__
        )


class OpcPackage(_RelatableMixin):
    """Main API class for |python-opc|.

    A new instance is constructed by calling the :meth:`open` classmethod with a path
    to a package file or file-like object containing a package (.pptx file).
    """

    def __init__(self, pkg_file):
        self._pkg_file = pkg_file

    @classmethod
    def open(cls, pkg_file):
        """Return an |OpcPackage| instance loaded with the contents of `pkg_file`."""
        return cls(pkg_file)._load()

    def drop_rel(self, rId):
        """Remove relationship identified by `rId`."""
        self._rels.pop(rId)

    def iter_parts(self):
        """Generate exactly one reference to each part in the package."""
        visited = set()
        for rel in self.iter_rels():
            if rel.is_external:
                continue
            part = rel.target_part
            if part in visited:
                continue
            yield part
            visited.add(part)

    def iter_rels(self):
        """Generate exactly one reference to each relationship in package.

        Performs a depth-first traversal of the rels graph.
        """
        visited = set()

        def walk_rels(rels):
            for rel in rels:
                yield rel
                # --- external items can have no relationships ---
                if rel.is_external:
                    continue
                # --- all relationships other than those for the package belong to a
                # --- part. Once that part has been processed, processing it again
                # --- would lead to the same relationships appearing more than once.
                part = rel.target_part
                if part in visited:
                    continue
                visited.add(part)
                # --- recurse into relationships of each unvisited target-part ---
                for rel in walk_rels(part.rels):
                    yield rel

        for rel in walk_rels(self._rels):
            yield rel

    @property
    def main_document_part(self):
        """Return |Part| subtype serving as the main document part for this package.

        In this case it will be a |Presentation| part.
        """
        return self.part_related_by(RT.OFFICE_DOCUMENT)

    def next_partname(self, tmpl):
        """Return |PackURI| next available partname matching `tmpl`.

        `tmpl` is a printf (%)-style template string containing a single replacement
        item, a '%d' to be used to insert the integer portion of the partname.
        Example: '/ppt/slides/slide%d.xml'
        """
        # --- expected next partname is tmpl % n where n is one greater than the number
        # --- of existing partnames that match tmpl. Speed up finding the next one
        # --- (maybe) by searching from the end downward rather than from 1 upward.
        prefix = tmpl[: (tmpl % 42).find("42")]
        partnames = set(
            p.partname for p in self.iter_parts() if p.partname.startswith(prefix)
        )
        for n in range(len(partnames) + 1, 0, -1):
            candidate_partname = tmpl % n
            if candidate_partname not in partnames:
                return PackURI(candidate_partname)
        raise Exception(  # pragma: no cover
            "ProgrammingError: ran out of candidate_partnames"
        )

    def save(self, pkg_file):
        """Save this package to `pkg_file`.

        `file` can be either a path to a file (a string) or a file-like object.
        """
        PackageWriter.write(pkg_file, self._rels, tuple(self.iter_parts()))

    def _load(self):
        """Return the package after loading all parts and relationships."""
        pkg_xml_rels, parts = _PackageLoader.load(self._pkg_file, self)
        self._rels.load_from_xml(PACKAGE_URI, pkg_xml_rels, parts)
        return self

    @lazyproperty
    def _rels(self):
        """|Relationships| object containing relationships of this package."""
        return _Relationships(PACKAGE_URI.baseURI)


class _PackageLoader(object):
    """Function-object that loads a package from disk (or other store)."""

    def __init__(self, pkg_file, package):
        self._pkg_file = pkg_file
        self._package = package

    @classmethod
    def load(cls, pkg_file, package):
        """Return (pkg_xml_rels, parts) pair resulting from loading `pkg_file`.

        The returned `parts` value is a {partname: part} mapping with each part in the
        package included and constructed complete with its relationships to other parts
        in the package.

        The returned `pkg_xml_rels` value is a `CT_Relationships` object containing the
        parsed package relationships. It is the caller's responsibility (the package
        object) to load those relationships into its |_Relationships| object.
        """
        return cls(pkg_file, package)._load()

    def _load(self):
        """Return (pkg_xml_rels, parts) pair resulting from loading pkg_file."""
        parts, xml_rels = self._parts, self._xml_rels

        for partname, part in parts.items():
            part.load_rels_from_xml(xml_rels[partname], parts)

        return xml_rels["/"], parts

    @lazyproperty
    def _content_types(self):
        """|_ContentTypeMap| object providing content-types for items of this package.

        Provides a content-type (MIME-type) for any given partname.
        """
        return _ContentTypeMap.from_xml(self._package_reader[CONTENT_TYPES_URI])

    @lazyproperty
    def _package_reader(self):
        """|PackageReader| object providing access to package-items in pkg_file."""
        return PackageReader(self._pkg_file)

    @lazyproperty
    def _parts(self):
        """dict {partname: Part} populated with parts loading from package.

        Among other duties, this collection is passed to each relationships collection
        so each relationship can resolve a reference to its target part when required.
        This reference can only be reliably carried out once the all parts have been
        loaded.
        """
        content_types = self._content_types
        package = self._package
        package_reader = self._package_reader

        return {
            partname: PartFactory(
                partname,
                content_types[partname],
                package,
                blob=package_reader[partname],
            )
            for partname in (p for p in self._xml_rels.keys() if p != "/")
            # --- invalid partnames can arise in some packages; ignore those rather
            # --- than raise an exception.
            if partname in package_reader
        }

    @lazyproperty
    def _xml_rels(self):
        """dict {partname: xml_rels} for package and all package parts.

        This is used as the basis for other loading operations such as loading parts and
        populating their relationships.
        """
        xml_rels = {}
        visited_partnames = set()

        def load_rels(source_partname, rels):
            """Populate `xml_rels` dict by traversing relationships depth-first."""
            xml_rels[source_partname] = rels
            visited_partnames.add(source_partname)
            base_uri = source_partname.baseURI

            # --- recursion stops when there are no unvisited partnames in rels ---
            for rel in rels:
                if rel.targetMode == RTM.EXTERNAL:
                    continue
                target_partname = PackURI.from_rel_ref(base_uri, rel.target_ref)
                if target_partname in visited_partnames:
                    continue
                load_rels(target_partname, self._xml_rels_for(target_partname))

        load_rels(PACKAGE_URI, self._xml_rels_for(PACKAGE_URI))
        return xml_rels

    def _xml_rels_for(self, partname):
        """Return CT_Relationships object formed by parsing rels XML for `partname`.

        A CT_Relationships object is returned in all cases. A part that has no
        relationships receives an "empty" CT_Relationships object, i.e. containing no
        `CT_Relationship` objects.
        """
        rels_xml = self._package_reader.rels_xml_for(partname)
        return CT_Relationships.new() if rels_xml is None else parse_xml(rels_xml)


class Part(_RelatableMixin):
    """Base class for package parts.

    Provides common properties and methods, but intended to be subclassed in client code
    to implement specific part behaviors. Also serves as the default class for parts
    that are not yet given specific behaviors.
    """

    def __init__(self, partname, content_type, package, blob=None):
        # --- XmlPart subtypes, don't store a blob (the original XML) ---
        self._partname = partname
        self._content_type = content_type
        self._package = package
        self._blob = blob

    @classmethod
    def load(cls, partname, content_type, package, blob):
        """Return `cls` instance loaded from arguments.

        This one is a straight pass-through, but subtypes may do some pre-processing,
        see XmlPart for an example.
        """
        return cls(partname, content_type, package, blob)

    @property
    def blob(self):
        """Contents of this package part as a sequence of bytes.

        May be text (XML generally) or binary. Intended to be overridden by subclasses.
        Default behavior is to return the blob initial loaded during `Package.open()`
        operation.
        """
        return self._blob

    @blob.setter
    def blob(self, bytes_):
        """Note that not all subclasses use the part blob as their blob source.

        In particular, the |XmlPart| subclass uses its `self._element` to serialize a
        blob on demand. This works fine for binary parts though.
        """
        self._blob = bytes_

    @lazyproperty
    def content_type(self):
        """Content-type (MIME-type) of this part."""
        return self._content_type

    def drop_rel(self, rId):
        """Remove relationship identified by `rId` if its reference count is under 2.

        Relationships with a reference count of 0 are implicit relationships. Note that
        only XML parts can drop relationships.
        """
        if self._rel_ref_count(rId) < 2:
            self._rels.pop(rId)

    def load_rels_from_xml(self, xml_rels, parts):
        """load _Relationships for this part from `xml_rels`.

        Part references are resolved using the `parts` dict that maps each partname to
        the loaded part with that partname. These relationships are loaded from a
        serialized package and so already have assigned rIds. This method is only used
        during package loading.
        """
        self._rels.load_from_xml(self._partname.baseURI, xml_rels, parts)

    @lazyproperty
    def package(self):
        """|OpcPackage| instance this part belongs to."""
        return self._package

    @property
    def partname(self):
        """|PackURI| partname for this part, e.g. "/ppt/slides/slide1.xml"."""
        return self._partname

    @partname.setter
    def partname(self, partname):
        if not isinstance(partname, PackURI):
            raise TypeError(  # pragma: no cover
                "partname must be instance of PackURI, got '%s'"
                % type(partname).__name__
            )
        self._partname = partname

    @lazyproperty
    def rels(self):
        """|Relationships| collection of relationships from this part to other parts."""
        # --- this must be public to allow the part graph to be traversed ---
        return self._rels

    def _blob_from_file(self, file):
        """Return bytes of `file`, which is either a str path or a file-like object."""
        # --- a str `file` is assumed to be a path ---
        if is_string(file):
            with open(file, "rb") as f:
                return f.read()

        # --- otherwise, assume `file` is a file-like object
        # --- reposition file cursor if it has one
        if callable(getattr(file, "seek")):
            file.seek(0)
        return file.read()

    def _rel_ref_count(self, rId):
        """Return int count of references in this part's XML to `rId`."""
        return len([r for r in self._element.xpath("//@r:id") if r == rId])

    @lazyproperty
    def _rels(self):
        """|Relationships| object containing relationships from this part to others."""
        return _Relationships(self._partname.baseURI)


class XmlPart(Part):
    """Base class for package parts containing an XML payload, which is most of them.

    Provides additional methods to the |Part| base class that take care of parsing and
    reserializing the XML payload and managing relationships to other parts.
    """

    def __init__(self, partname, content_type, package, element):
        super(XmlPart, self).__init__(partname, content_type, package)
        self._element = element

    @classmethod
    def load(cls, partname, content_type, package, blob):
        """Return instance of `cls` loaded with parsed XML from `blob`."""
        return cls(partname, content_type, package, element=parse_xml(blob))

    @property
    def blob(self):
        """bytes XML serialization of this part."""
        return serialize_part_xml(self._element)

    @property
    def part(self):
        """This part.

        This is part of the parent protocol, "children" of the document will not know
        the part that contains them so must ask their parent object. That chain of
        delegation ends here for child objects.
        """
        return self


class PartFactory(object):
    """Constructs a registered subtype of |Part|.

    Client code can register a subclass of |Part| to be used for a package blob based on
    its content type.
    """

    part_type_for = {}

    def __new__(cls, partname, content_type, package, blob):
        PartClass = cls._part_cls_for(content_type)
        return PartClass.load(partname, content_type, package, blob)

    @classmethod
    def _part_cls_for(cls, content_type):
        """Return the custom part class registered for `content_type`.

        Returns |Part| if no custom class is registered for `content_type`.
        """
        if content_type in cls.part_type_for:
            return cls.part_type_for[content_type]
        return Part


class _ContentTypeMap(object):
    """Value type providing dict semantics for looking up content type by partname."""

    def __init__(self, overrides, defaults):
        self._overrides = overrides
        self._defaults = defaults

    def __getitem__(self, partname):
        """Return content-type (MIME-type) for part identified by *partname*."""
        if not isinstance(partname, PackURI):
            raise TypeError(
                "_ContentTypeMap key must be <type 'PackURI'>, got %s"
                % type(partname).__name__
            )

        if partname in self._overrides:
            return self._overrides[partname]

        if partname.ext in self._defaults:
            return self._defaults[partname.ext]

        raise KeyError(
            "no content-type for partname '%s' in [Content_Types].xml" % partname
        )

    @classmethod
    def from_xml(cls, content_types_xml):
        """Return |_ContentTypeMap| instance populated from `content_types_xml`."""
        types_elm = parse_xml(content_types_xml)
        overrides = CaseInsensitiveDict(
            (o.partName.lower(), o.contentType) for o in types_elm.override_lst
        )
        defaults = CaseInsensitiveDict(
            (d.extension.lower(), d.contentType) for d in types_elm.default_lst
        )
        return cls(overrides, defaults)


class _Relationships(Mapping):
    """Collection of |_Relationship| instances, largely having dict semantics.

    Relationships are keyed by their rId, but may also be found in other ways, such as
    by their relationship type. `rels` is a dict of |Relationship| objects keyed by
    their rId.

    Note that iterating this collection generates |Relationship| references (values),
    not rIds (keys) as it would for a dict.
    """

    def __init__(self, base_uri):
        self._base_uri = base_uri

    def __contains__(self, rId):
        """Implement 'in' operation, like `"rId7" in relationships`."""
        return rId in self._rels

    def __getitem__(self, rId):
        """Implement relationship lookup by rId using indexed access, like rels[rId]."""
        try:
            return self._rels[rId]
        except KeyError:
            raise KeyError("no relationship with key '%s'" % rId)

    def __iter__(self):
        """Implement iteration of relationships."""
        rels = self._rels
        return (rels[rId] for rId in sorted(rels.keys()))

    def __len__(self):
        """Return count of relationships in collection."""
        return len(self._rels)

    def get_or_add(self, reltype, target_part):
        """Return str rId of `reltype` to `target_part`.

        The rId of an existing matching relationship is used if present. Otherwise, a
        new relationship is added and that rId is returned.
        """
        existing_rId = self._get_matching(reltype, target_part)
        return (
            self._add_relationship(reltype, target_part)
            if existing_rId is None
            else existing_rId
        )

    def get_or_add_ext_rel(self, reltype, target_ref):
        """Return str rId of external relationship of `reltype` to `target_ref`.

        The rId of an existing matching relationship is used if present. Otherwise, a
        new relationship is added and that rId is returned.
        """
        existing_rId = self._get_matching(reltype, target_ref, is_external=True)
        return (
            self._add_relationship(reltype, target_ref, is_external=True)
            if existing_rId is None
            else existing_rId
        )

    def load_from_xml(self, base_uri, xml_rels, parts):
        """Replace any relationships in this collection with those from `xml_rels`."""

        def iter_valid_rels():
            """Filter out broken relationships such as those pointing to NULL."""
            for rel_elm in xml_rels.relationship_lst:
                # --- Occasionally a PowerPoint plugin or other client will "remove"
                # --- a relationship simply by "voiding" its Target value, like making
                # --- it "/ppt/slides/NULL". Skip any relationships linking to a
                # --- partname that is not present in the package.
                if rel_elm.targetMode == RTM.INTERNAL:
                    partname = PackURI.from_rel_ref(base_uri, rel_elm.target_ref)
                    if partname not in parts:
                        continue
                yield _Relationship.from_xml(base_uri, rel_elm, parts)

        self._rels.clear()
        self._rels.update((rel.rId, rel) for rel in iter_valid_rels())

    def part_with_reltype(self, reltype):
        """Return target part of relationship with matching `reltype`.

        Raises |KeyError| if not found and |ValueError| if more than one matching
        relationship is found.
        """
        rels_of_reltype = self._rels_by_reltype[reltype]

        if len(rels_of_reltype) == 0:
            raise KeyError("no relationship of type '%s' in collection" % reltype)

        if len(rels_of_reltype) > 1:
            raise ValueError(
                "multiple relationships of type '%s' in collection" % reltype
            )

        return rels_of_reltype[0].target_part

    def pop(self, rId):
        """Return |Relationship| identified by `rId` after removing it from collection.

        The caller is responsible for ensuring it is no longer required.
        """
        return self._rels.pop(rId)

    @property
    def xml(self):
        """bytes XML serialization of this relationship collection.

        This value is suitable for storage as a .rels file in an OPC package. Includes
        a `<?xml` header with encoding as UTF-8.
        """
        rels_elm = CT_Relationships.new()
        for rel in self:
            rels_elm.add_rel(rel.rId, rel.reltype, rel.target_ref, rel.is_external)
        return rels_elm.xml

    def _add_relationship(self, reltype, target, is_external=False):
        """Return str rId of |_Relationship| newly added to spec."""
        rId = self._next_rId
        self._rels[rId] = _Relationship(
            self._base_uri,
            rId,
            reltype,
            target_mode=RTM.EXTERNAL if is_external else RTM.INTERNAL,
            target=target,
        )
        return rId

    def _get_matching(self, reltype, target, is_external=False):
        """Return optional str rId of rel of `reltype`, `target`, and `is_external`.

        Returns `None` on no matching relationship
        """
        for rel in self._rels_by_reltype[reltype]:
            if rel.is_external != is_external:
                continue
            rel_target = rel.target_ref if rel.is_external else rel.target_part
            if rel_target != target:
                continue
            return rel.rId

        return None

    @property
    def _next_rId(self):
        """Next str rId available in collection.

        The next rId is the first unused key starting from "rId1" and making use of any
        gaps in numbering, e.g. 'rId2' for rIds ['rId1', 'rId3'].
        """
        # --- The common case is where all sequential numbers starting at "rId1" are
        # --- used and the next available rId is "rId%d" % (len(rels)+1). So we start
        # --- there and count down to produce the best performance.
        for n in range(len(self) + 1, 0, -1):
            rId_candidate = "rId%d" % n  # like 'rId19'
            if rId_candidate not in self._rels:
                return rId_candidate

    @lazyproperty
    def _rels(self):
        """dict {rId: _Relationship} containing relationships of this collection."""
        return dict()

    @property
    def _rels_by_reltype(self):
        """defaultdict {reltype: [rels]} for all relationships in collection."""
        D = collections.defaultdict(list)
        for rel in self:
            D[rel.reltype].append(rel)
        return D


class _Relationship(object):
    """Value object describing link from a part or package to another part."""

    def __init__(self, base_uri, rId, reltype, target_mode, target):
        self._base_uri = base_uri
        self._rId = rId
        self._reltype = reltype
        self._target_mode = target_mode
        self._target = target

    @classmethod
    def from_xml(cls, base_uri, rel, parts):
        """Return |_Relationship| object based on CT_Relationship element `rel`."""
        target = (
            rel.target_ref
            if rel.targetMode == RTM.EXTERNAL
            else parts[PackURI.from_rel_ref(base_uri, rel.target_ref)]
        )
        return cls(base_uri, rel.rId, rel.reltype, rel.targetMode, target)

    @lazyproperty
    def is_external(self):
        """True if target_mode is `RTM.EXTERNAL`.

        An external relationship is a link to a resource outside the package, such as
        a web-resource (URL).
        """
        return self._target_mode == RTM.EXTERNAL

    @lazyproperty
    def reltype(self):
        """Member of RELATIONSHIP_TYPE describing relationship of target to source."""
        return self._reltype

    @lazyproperty
    def rId(self):
        """str relationship-id, like 'rId9'.

        Corresponds to the `Id` attribute on the `CT_Relationship` element and
        uniquely identifies this relationship within its peers for the source-part or
        package.
        """
        return self._rId

    @lazyproperty
    def target_part(self):
        """|Part| or subtype referred to by this relationship."""
        if self.is_external:
            raise ValueError(
                "`.target_part` property on _Relationship is undefined when "
                "target-mode is external"
            )
        return self._target

    @lazyproperty
    def target_partname(self):
        """|PackURI| instance containing partname targeted by this relationship.

        Raises `ValueError` on reference if target_mode is external. Use
        :attr:`target_mode` to check before referencing.
        """
        if self.is_external:
            raise ValueError(
                "`.target_partname` property on _Relationship is undefined when "
                "target-mode is external"
            )
        return self._target.partname

    @lazyproperty
    def target_ref(self):
        """str reference to relationship target.

        For internal relationships this is the relative partname, suitable for
        serialization purposes. For an external relationship it is typically a URL.
        """
        return (
            self._target
            if self.is_external
            else self.target_partname.relative_ref(self._base_uri)
        )



# encoding: utf-8

"""Unit test suite for `pptx.parts.media` module."""

from pptx.media import Video
from pptx.package import Package
from pptx.parts.media import MediaPart

from ..unitutil.mock import initializer_mock, instance_mock


class DescribeMediaPart(object):
    """Unit-test suite for `pptx.parts.media.MediaPart` objects."""

    def it_can_construct_from_a_media_object(self, request):
        media_ = instance_mock(request, Video)
        _init_ = initializer_mock(request, MediaPart)
        package_ = instance_mock(request, Package)
        package_.next_media_partname.return_value = "media42.mp4"
        media_.blob, media_.content_type = b"blob-bytes", "video/mp4"

        media_part = MediaPart.new(package_, media_)

        package_.next_media_partname.assert_called_once_with(media_.ext)
        _init_.assert_called_once_with(
            media_part, "media42.mp4", media_.content_type, package_, media_.blob
        )
        assert isinstance(media_part, MediaPart)

    def it_knows_the_sha1_hash_of_the_media(self):
        assert MediaPart(None, None, None, b"blobish-bytes").sha1 == (
            "61efc464c21e54cfc1382fb5b6ef7512e141ceae"
        )



# encoding: utf-8

"""Utility functions wrapping the excellent `mock` library."""

from __future__ import absolute_import

import sys

if sys.version_info >= (3, 3):
    from unittest import mock  # noqa
    from unittest.mock import ANY, call, MagicMock  # noqa
    from unittest.mock import create_autospec, Mock, mock_open, patch, PropertyMock
else:  # pragma: no cover
    import mock  # noqa
    from mock import ANY, call, MagicMock  # noqa
    from mock import create_autospec, Mock, mock_open, patch, PropertyMock


def class_mock(request, q_class_name, autospec=True, **kwargs):
    """Return a mock patching the class with qualified name *q_class_name*.

    The mock is autospec'ed based on the patched class unless the optional argument
    *autospec* is set to False. Any other keyword arguments are passed through to
    Mock(). Patch is reversed after calling test returns.
    """
    _patch = patch(q_class_name, autospec=autospec, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def cls_attr_mock(request, cls, attr_name, name=None, **kwargs):  # pragma: no cover
    """Return a mock for attribute (class variable) `attr_name` on `cls`.

    Patch is reversed after pytest uses it.
    """
    name = request.fixturename if name is None else name
    _patch = patch.object(cls, attr_name, name=name, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def function_mock(request, q_function_name, autospec=True, **kwargs):
    """Return mock patching function with qualified name `q_function_name`.

    Patch is reversed after calling test returns.
    """
    _patch = patch(q_function_name, autospec=autospec, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def initializer_mock(request, cls, autospec=True, **kwargs):
    """Return mock for __init__() method on `cls`.

    The patch is reversed after pytest uses it.
    """
    _patch = patch.object(
        cls, "__init__", autospec=autospec, return_value=None, **kwargs
    )
    request.addfinalizer(_patch.stop)
    return _patch.start()


def instance_mock(request, cls, name=None, spec_set=True, **kwargs):
    """Return mock for instance of `cls` that draws its spec from that class.

    The mock does not allow new attributes to be set on the instance. If `name` is
    missing or |None|, the name of the returned |Mock| instance is set to
    `request.fixturename`. Additional keyword arguments are passed through to the Mock()
    call that creates the mock.
    """
    name = name if name is not None else request.fixturename
    return create_autospec(cls, _name=name, spec_set=spec_set, instance=True, **kwargs)


def loose_mock(request, name=None, **kwargs):
    """Return a "loose" mock, meaning it has no spec to constrain calls on it.

    Additional keyword arguments are passed through to Mock(). If called without a name,
    it is assigned the name of the fixture.
    """
    return Mock(name=request.fixturename if name is None else name, **kwargs)


def method_mock(request, cls, method_name, autospec=True, **kwargs):
    """Return mock for method `method_name` on `cls`.

    The patch is reversed after pytest uses it.
    """
    _patch = patch.object(cls, method_name, autospec=autospec, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def open_mock(request, module_name, **kwargs):
    """Return a mock for the builtin `open()` method in `module_name`."""
    target = "%s.open" % module_name
    _patch = patch(target, mock_open(), create=True, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def property_mock(request, cls, prop_name, **kwargs):
    """Return a mock for property `prop_name` on class `cls`."""
    _patch = patch.object(cls, prop_name, new_callable=PropertyMock, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def var_mock(request, q_var_name, **kwargs):
    """Return mock patching the variable with qualified name *q_var_name*."""
    _patch = patch(q_var_name, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()



# encoding: utf-8

"""Unit-test suite for the `pptx.opc.packuri` module."""

import pytest

from pptx.opc.packuri import PackURI


class DescribePackURI(object):
    """Unit-test suite for the `pptx.opc.packuri.PackURI` objects."""

    def it_can_construct_from_relative_ref(self):
        pack_uri = PackURI.from_rel_ref(
            "/ppt/slides", "../slideLayouts/slideLayout1.xml"
        )
        assert pack_uri == "/ppt/slideLayouts/slideLayout1.xml"

    def it_should_raise_on_construct_with_bad_pack_uri_str(self):
        with pytest.raises(ValueError):
            PackURI("foobar")

    @pytest.mark.parametrize(
        "uri, expected_value",
        (
            ("/", "/"),
            ("/ppt/presentation.xml", "/ppt"),
            ("/ppt/slides/slide1.xml", "/ppt/slides"),
        ),
    )
    def it_knows_its_base_URI(self, uri, expected_value):
        assert PackURI(uri).baseURI == expected_value

    @pytest.mark.parametrize(
        "uri, expected_value",
        (
            ("/", ""),
            ("/ppt/presentation.xml", "xml"),
            ("/ppt/media/image.PnG", "PnG"),
        ),
    )
    def it_knows_its_extension(self, uri, expected_value):
        assert PackURI(uri).ext == expected_value

    @pytest.mark.parametrize(
        "uri, expected_value",
        (
            ("/", ""),
            ("/ppt/presentation.xml", "presentation.xml"),
            ("/ppt/media/image.png", "image.png"),
        ),
    )
    def it_knows_its_filename(self, uri, expected_value):
        assert PackURI(uri).filename == expected_value

    @pytest.mark.parametrize(
        "uri, expected_value",
        (
            ("/", None),
            ("/ppt/presentation.xml", None),
            ("/ppt/,foo,grob!.xml", None),
            ("/ppt/media/image42.png", 42),
        ),
    )
    def it_knows_the_filename_index(self, uri, expected_value):
        assert PackURI(uri).idx == expected_value

    @pytest.mark.parametrize(
        "uri, base_uri, expected_value",
        (
            ("/ppt/presentation.xml", "/", "ppt/presentation.xml"),
            (
                "/ppt/slideMasters/slideMaster1.xml",
                "/ppt",
                "slideMasters/slideMaster1.xml",
            ),
            (
                "/ppt/slideLayouts/slideLayout1.xml",
                "/ppt/slides",
                "../slideLayouts/slideLayout1.xml",
            ),
        ),
    )
    def it_can_compute_its_relative_reference(self, uri, base_uri, expected_value):
        assert PackURI(uri).relative_ref(base_uri) == expected_value

    @pytest.mark.parametrize(
        "uri, expected_value",
        (
            ("/", "/_rels/.rels"),
            ("/ppt/presentation.xml", "/ppt/_rels/presentation.xml.rels"),
            ("/ppt/slides/slide42.xml", "/ppt/slides/_rels/slide42.xml.rels"),
        ),
    )
    def it_knows_the_uri_of_its_rels_part(self, uri, expected_value):
        assert PackURI(uri).rels_uri == expected_value



# encoding: utf-8

"""Unit-test suite for `pptx.enum` subpackage, focused on base classes.

Configured a little differently because of the meta-programming, the two enumeration
classes at the top constitute the entire fixture and most of the tests themselves just
make assertions on those.
"""

import pytest

from pptx.enum.base import (
    alias,
    Enumeration,
    EnumMember,
    ReturnValueOnlyEnumMember,
    XmlEnumeration,
    XmlMappedEnumMember,
)
from pptx.enum.shapes import PROG_ID, _ProgIdEnum
from pptx.util import Emu


@alias("BARFOO")
class FOOBAR(Enumeration):
    """
    Enumeration docstring
    """

    __ms_name__ = "MsoFoobar"

    __url__ = "http://msdn.microsoft.com/foobar.aspx"

    __members__ = (
        EnumMember("READ_WRITE", 1, "Readable and settable"),
        ReturnValueOnlyEnumMember("READ_ONLY", -2, "Return value only"),
    )


@alias("XML-FU")
class XMLFOO(XmlEnumeration):
    """
    XmlEnumeration docstring
    """

    __ms_name__ = "MsoXmlFoobar"

    __url__ = "http://msdn.microsoft.com/msoxmlfoobar.aspx"

    __members__ = (
        XmlMappedEnumMember(None, None, None, "No setting"),
        XmlMappedEnumMember("XML_RW", 42, "attrVal", "Read/write setting"),
        ReturnValueOnlyEnumMember("RO", -2, "Return value only;"),
    )


class DescribeEnumeration(object):
    def it_has_the_right_metaclass(self):
        assert type(FOOBAR).__name__ == "MetaEnumeration"

    def it_provides_an_EnumValue_instance_for_each_named_member(self):
        for obj in (FOOBAR.READ_WRITE, FOOBAR.READ_ONLY):
            assert type(obj).__name__ == "EnumValue"

    def it_provides_the_enumeration_value_for_each_named_member(self):
        assert FOOBAR.READ_WRITE == 1
        assert FOOBAR.READ_ONLY == -2

    def it_knows_if_a_setting_is_valid(self):
        FOOBAR.validate(FOOBAR.READ_WRITE)
        with pytest.raises(ValueError):
            FOOBAR.validate("foobar")
        with pytest.raises(ValueError):
            FOOBAR.validate(FOOBAR.READ_ONLY)

    def it_can_be_referred_to_by_a_convenience_alias_if_defined(self):
        assert BARFOO is FOOBAR  # noqa


class DescribeEnumValue(object):
    def it_provides_its_symbolic_name_as_its_string_value(self):
        assert ("%s" % FOOBAR.READ_WRITE) == "READ_WRITE (1)"

    def it_provides_its_description_as_its_docstring(self):
        assert FOOBAR.READ_ONLY.__doc__ == "Return value only"


class DescribeXmlEnumeration(object):
    def it_knows_the_XML_value_for_each_of_its_xml_members(self):
        assert XMLFOO.to_xml(XMLFOO.XML_RW) == "attrVal"
        assert XMLFOO.to_xml(42) == "attrVal"
        with pytest.raises(ValueError):
            XMLFOO.to_xml(XMLFOO.RO)

    def it_can_map_each_of_its_xml_members_from_the_XML_value(self):
        assert XMLFOO.from_xml(None) is None
        assert XMLFOO.from_xml("attrVal") == XMLFOO.XML_RW
        assert str(XMLFOO.from_xml("attrVal")) == "XML_RW (42)"


class Describe_ProgIdEnum(object):
    """Unit-test suite for `pptx.enum.shapes._ProgIdEnum."""

    def it_provides_access_to_its_members(self):
        assert type(PROG_ID.XLSX) == _ProgIdEnum.Member

    def it_can_test_an_item_for_membership(self):
        assert PROG_ID.XLSX in PROG_ID

    def it_has_a_readable_representation_for_itself(self):
        assert repr(PROG_ID) == "pptx.enum.shapes.PROG_ID"

    def it_has_a_readable_representation_for_each_of_its_members(self):
        assert repr(PROG_ID.XLSX) == "PROG_ID.XLSX"

    def it_has_attributes_on_each_member(self):
        assert PROG_ID.XLSX.height == Emu(609600)
        assert PROG_ID.XLSX.icon_filename == "xlsx-icon.emf"
        assert PROG_ID.XLSX.progId == "Excel.Sheet.12"
        assert PROG_ID.XLSX.width == Emu(965200)
